"""
Common Utilities
================

Defines the common algebra utilities objects that don't fall in any specific
category.
"""

from __future__ import annotations

import functools
import numpy as np

from colour.hints import (
    Any,
    ArrayLike,
    Boolean,
    Callable,
    Floating,
    FloatingOrArrayLike,
    FloatingOrNDArray,
    Integer,
    NDArray,
    Optional,
)
from colour.utilities import as_float_array, as_float, tsplit

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "is_spow_enabled",
    "set_spow_enable",
    "spow_enable",
    "spow",
    "normalise_maximum",
    "vector_dot",
    "matrix_dot",
    "linear_conversion",
    "linstep_function",
    "lerp",
    "smoothstep_function",
    "smooth",
    "is_identity",
]

# TODO: Annotate with "bool" when Python 3.7 is dropped.
_SPOW_ENABLED = True
"""
Global variable storing the current *Colour* safe / symmetrical power function
enabled state.
"""


def is_spow_enabled() -> bool:
    """
    Return whether *Colour* safe / symmetrical power function is enabled.

    Returns
    -------
    :class:`bool`
        Whether *Colour* safe / symmetrical power function is enabled.

    Examples
    --------
    >>> with spow_enable(False):
    ...     is_spow_enabled()
    False
    >>> with spow_enable(True):
    ...     is_spow_enabled()
    True
    """

    return _SPOW_ENABLED


def set_spow_enable(enable: bool):
    """
    Set *Colour* safe / symmetrical power function enabled state.

    Parameters
    ----------
    enable
        Whether to enable *Colour* safe / symmetrical power function.

    Examples
    --------
    >>> with spow_enable(is_spow_enabled()):
    ...     print(is_spow_enabled())
    ...     set_spow_enable(False)
    ...     print(is_spow_enabled())
    True
    False
    """

    global _SPOW_ENABLED

    _SPOW_ENABLED = enable


class spow_enable:
    """
    Define a context manager and decorator temporarily setting *Colour* safe /
    symmetrical power function enabled state.

    Parameters
    ----------
    enable
        Whether to enable or disable *Colour* safe / symmetrical power
        function.
    """

    def __init__(self, enable: bool):
        self._enable = enable
        self._previous_state = is_spow_enabled()

    def __enter__(self) -> spow_enable:
        """
        Set the *Colour* safe / symmetrical power function enabled state
        upon entering the context manager.
        """

        set_spow_enable(self._enable)

        return self

    def __exit__(self, *args: Any):
        """
        Set the *Colour* safe / symmetrical power function enabled state
        upon exiting the context manager.
        """

        set_spow_enable(self._previous_state)

    def __call__(self, function: Callable) -> Callable:
        """Call the wrapped definition."""

        @functools.wraps(function)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            with self:
                return function(*args, **kwargs)

        return wrapper


def spow(a: FloatingOrArrayLike, p: FloatingOrArrayLike) -> FloatingOrNDArray:
    """
    Raise given array :math:`a` to the power :math:`p` as follows:
    :math:`sign(a) * |a|^p`.

    This definition avoids NaNs generation when array :math:`a` is negative and
    the power :math:`p` is fractional. This behaviour can be enabled or
    disabled with the :func:`colour.algebra.set_spow_enable` definition or with
    the :func:`spow_enable` context manager.

    Parameters
    ----------
    a
        Array :math:`a`.
    p
        Power :math:`p`.

    Returns
    -------
    :class:`np.floating` or :class:`numpy.ndarray`
        Array :math:`a` safely raised to the power :math:`p`.

    Examples
    --------
    >>> np.power(-2, 0.15)
    nan
    >>> spow(-2, 0.15)  # doctest: +ELLIPSIS
    -1.1095694...
    >>> spow(0, 0)
    0.0
    """

    if not _SPOW_ENABLED:
        return np.power(a, p)

    a = np.atleast_1d(a)
    p = as_float_array(p)

    a_p = np.sign(a) * np.abs(a) ** p

    a_p[np.isnan(a_p)] = 0

    return as_float(a_p)


def normalise_maximum(
    a: ArrayLike,
    axis: Optional[Integer] = None,
    factor: Floating = 1,
    clip: Boolean = True,
) -> NDArray:
    """
    Normalise given array :math:`a` values by :math:`a` maximum value and
    optionally clip them between.

    Parameters
    ----------
    a
        Array :math:`a` to normalise.
    axis
        Normalization axis.
    factor
        Normalization factor.
    clip
        Clip values to domain [0, 'factor'].

    Returns
    -------
    :class:`numpy.ndarray`
        Maximum normalised array :math:`a`.

    Examples
    --------
    >>> a = np.array([0.48222001, 0.31654775, 0.22070353])
    >>> normalise_maximum(a)  # doctest: +ELLIPSIS
    array([ 1.        ,  0.6564384...,  0.4576822...])
    """

    a = as_float_array(a)

    maximum = np.max(a, axis=axis)
    a = a * (1 / maximum[..., np.newaxis]) * factor

    return np.clip(a, 0, factor) if clip else a


def vector_dot(m: ArrayLike, v: ArrayLike) -> NDArray:
    """
    Perform the dot product of the matrix array :math:`m` with the vector
    array :math:`v`.

    This definition is a convenient wrapper around :func:`np.einsum` with the
    following subscripts: *'...ij,...j->...i'*.

    Parameters
    ----------
    m
        Matrix array :math:`m`.
    v
        Vector array :math:`v`.

    Returns
    -------
    :class:`numpy.ndarray`
        Transformed vector array :math:`v`.

    Examples
    --------
    >>> m = np.array(
    ...     [[0.7328, 0.4296, -0.1624],
    ...      [-0.7036, 1.6975, 0.0061],
    ...      [0.0030, 0.0136, 0.9834]]
    ... )
    >>> m = np.reshape(np.tile(m, (6, 1)), (6, 3, 3))
    >>> v = np.array([0.20654008, 0.12197225, 0.05136952])
    >>> v = np.tile(v, (6, 1))
    >>> vector_dot(m, v)  # doctest: +ELLIPSIS
    array([[ 0.1954094...,  0.0620396...,  0.0527952...],
           [ 0.1954094...,  0.0620396...,  0.0527952...],
           [ 0.1954094...,  0.0620396...,  0.0527952...],
           [ 0.1954094...,  0.0620396...,  0.0527952...],
           [ 0.1954094...,  0.0620396...,  0.0527952...],
           [ 0.1954094...,  0.0620396...,  0.0527952...]])
    """

    return np.einsum("...ij,...j->...i", as_float_array(m), as_float_array(v))


def matrix_dot(a: ArrayLike, b: ArrayLike) -> NDArray:
    """
    Perform the dot product of the matrix array :math:`a` with the matrix
    array :math:`b`.

    This definition is a convenient wrapper around :func:`np.einsum` with the
    following subscripts: *'...ij,...jk->...ik'*.

    Parameters
    ----------
    a
        Matrix array :math:`a`.
    b
        Matrix array :math:`b`.

    Returns
    -------
    :class:`numpy.ndarray`

    Examples
    --------
    >>> a = np.array(
    ...     [[0.7328, 0.4296, -0.1624],
    ...      [-0.7036, 1.6975, 0.0061],
    ...      [0.0030, 0.0136, 0.9834]]
    ... )
    >>> a = np.reshape(np.tile(a, (6, 1)), (6, 3, 3))
    >>> b = a
    >>> matrix_dot(a, b)  # doctest: +ELLIPSIS
    array([[[ 0.2342420...,  1.0418482..., -0.2760903...],
            [-1.7099407...,  2.5793226...,  0.1306181...],
            [-0.0044203...,  0.0377490...,  0.9666713...]],
    <BLANKLINE>
           [[ 0.2342420...,  1.0418482..., -0.2760903...],
            [-1.7099407...,  2.5793226...,  0.1306181...],
            [-0.0044203...,  0.0377490...,  0.9666713...]],
    <BLANKLINE>
           [[ 0.2342420...,  1.0418482..., -0.2760903...],
            [-1.7099407...,  2.5793226...,  0.1306181...],
            [-0.0044203...,  0.0377490...,  0.9666713...]],
    <BLANKLINE>
           [[ 0.2342420...,  1.0418482..., -0.2760903...],
            [-1.7099407...,  2.5793226...,  0.1306181...],
            [-0.0044203...,  0.0377490...,  0.9666713...]],
    <BLANKLINE>
           [[ 0.2342420...,  1.0418482..., -0.2760903...],
            [-1.7099407...,  2.5793226...,  0.1306181...],
            [-0.0044203...,  0.0377490...,  0.9666713...]],
    <BLANKLINE>
           [[ 0.2342420...,  1.0418482..., -0.2760903...],
            [-1.7099407...,  2.5793226...,  0.1306181...],
            [-0.0044203...,  0.0377490...,  0.9666713...]]])
    """

    return np.einsum(
        "...ij,...jk->...ik", as_float_array(a), as_float_array(b)
    )


def linear_conversion(
    a: ArrayLike, old_range: ArrayLike, new_range: ArrayLike
) -> NDArray:
    """
    Perform a simple linear conversion of given array :math:`a` between the
    old and new ranges.

    Parameters
    ----------
    a
        Array :math:`a` to perform the linear conversion onto.
    old_range
        Old range.
    new_range
        New range.

    Returns
    -------
    :class:`numpy.ndarray`
        Linear conversion result.

    Examples
    --------
    >>> a = np.linspace(0, 1, 10)
    >>> linear_conversion(a, np.array([0, 1]), np.array([1, 10]))
    array([  1.,   2.,   3.,   4.,   5.,   6.,   7.,   8.,   9.,  10.])
    """

    a = as_float_array(a)

    in_min, in_max = tsplit(old_range)
    out_min, out_max = tsplit(new_range)

    return ((a - in_min) / (in_max - in_min)) * (out_max - out_min) + out_min


def linstep_function(
    x: FloatingOrArrayLike,
    a: FloatingOrArrayLike = 0,
    b: FloatingOrArrayLike = 1,
    clip: Boolean = False,
) -> NDArray:
    """
    Perform a simple linear interpolation between given array :math:`a` and
    array :math:`b` using :math:`x` array.

    Parameters
    ----------
    x
        Array :math:`x` value to use to interpolate between array :math:`a` and
        array :math:`b`.
    a
        Array :math:`a`, the start of the range in which to interpolate.
    b
        Array :math:`b`, the end of the range in which to interpolate.
    clip
        Whether to clip the output values to range [``a``, ``b``].

    Returns
    -------
    :class:`numpy.ndarray`
        Linear interpolation result.

    Examples
    --------
    >>> a = 0
    >>> b = 2
    >>> linstep_function(0.5, a, b)
    1.0
    """

    x = as_float_array(x)
    a = as_float_array(a)
    b = as_float_array(b)

    y = (1 - x) * a + x * b

    return np.clip(y, a, b) if clip else y


lerp = linstep_function


def smoothstep_function(
    x: FloatingOrArrayLike,
    a: FloatingOrArrayLike = 0,
    b: FloatingOrArrayLike = 1,
    clip: Boolean = False,
) -> NDArray:
    """
    Evaluate the *smoothstep* sigmoid-like function on array :math:`x`.

    Parameters
    ----------
    x
        Array :math:`x`.
    a
        Low input domain limit, i.e. the left edge.
    b
        High input domain limit, i.e. the right edge.
    clip
        Whether to scale, bias and clip input values to domain [``a``, ``b``].

    Returns
    -------
    :class:`numpy.ndarray`
        Array :math:`x` after *smoothstep* sigmoid-like function evaluation.

    Examples
    --------
    >>> x = np.linspace(-2, 2, 5)
    >>> smoothstep_function(x, -2, 2, clip=True)
    array([ 0.     ,  0.15625,  0.5    ,  0.84375,  1.     ])
    """

    x = as_float_array(x)
    a = as_float_array(a)
    b = as_float_array(b)

    i = np.clip((x - a) / (b - a), 0, 1) if clip else x

    return (i**2) * (3 - 2 * i)


smooth = smoothstep_function


def is_identity(a: ArrayLike) -> Boolean:
    """
    Return whether :math:`a` array is an identity matrix.

    Parameters
    ----------
    a
        Array :math:`a` to test.

    Returns
    -------
    :class:`bool`
        Whether :math:`a` array is an identity matrix.

    Examples
    --------
    >>> is_identity(np.array([1, 0, 0, 0, 1, 0, 0, 0, 1]).reshape(3, 3))
    True
    >>> is_identity(np.array([1, 2, 0, 0, 1, 0, 0, 0, 1]).reshape(3, 3))
    False
    """

    return np.array_equal(np.identity(len(np.diag(a))), a)



"""Defines the unit tests for the :mod:`colour.models.jzazbz` module."""

import numpy as np
import unittest
from itertools import permutations

from colour.models import (
    XYZ_to_Izazbz,
    Izazbz_to_XYZ,
    XYZ_to_Jzazbz,
    Jzazbz_to_XYZ,
)
from colour.utilities import domain_range_scale, ignore_numpy_errors

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestXYZ_to_Izazbz",
    "TestIzazbz_to_XYZ",
    "TestXYZ_to_Jzazbz",
    "TestJzazbz_to_XYZ",
]


class TestXYZ_to_Izazbz(unittest.TestCase):
    """
    Define :func:`colour.models.jzazbz.TestXYZ_to_Izazbz` definition unit
    tests methods.
    """

    def test_XYZ_to_Izazbz(self):
        """Test :func:`colour.models.jzazbz.XYZ_to_Izazbz` definition."""

        np.testing.assert_almost_equal(
            XYZ_to_Izazbz(np.array([0.20654008, 0.12197225, 0.05136952])),
            np.array([0.01207793, 0.00924302, 0.00526007]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            XYZ_to_Izazbz(np.array([0.14222010, 0.23042768, 0.10495772])),
            np.array([0.01397346, -0.00608426, 0.00534077]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            XYZ_to_Izazbz(np.array([0.96907232, 1.00000000, 1.12179215])),
            np.array([0.03927203, 0.00064174, -0.00052906]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            XYZ_to_Izazbz(
                np.array([0.20654008, 0.12197225, 0.05136952]),
                method="Safdar 2021",
            ),
            np.array([0.01049146, 0.00924302, 0.00526007]),
            decimal=7,
        )

        np.testing.assert_array_equal(
            XYZ_to_Izazbz(
                np.array([0.20654008, 0.12197225, 0.05136952]),
                method="Safdar 2021",
            ),
            XYZ_to_Izazbz(
                np.array([0.20654008, 0.12197225, 0.05136952]), method="ZCAM"
            ),
        )

    def test_n_dimensional_XYZ_to_Izazbz(self):
        """
        Test :func:`colour.models.jzazbz.XYZ_to_Izazbz` definition
        n-dimensional support.
        """

        XYZ = np.array([0.20654008, 0.12197225, 0.05136952])
        Izazbz = XYZ_to_Izazbz(XYZ)

        XYZ = np.tile(XYZ, (6, 1))
        Izazbz = np.tile(Izazbz, (6, 1))
        np.testing.assert_almost_equal(XYZ_to_Izazbz(XYZ), Izazbz, decimal=7)

        XYZ = np.reshape(XYZ, (2, 3, 3))
        Izazbz = np.reshape(Izazbz, (2, 3, 3))
        np.testing.assert_almost_equal(XYZ_to_Izazbz(XYZ), Izazbz, decimal=7)

    def test_domain_range_scale_XYZ_to_Izazbz(self):
        """
        Test :func:`colour.models.jzazbz.XYZ_to_Izazbz` definition domain and
        range scale support.
        """

        XYZ = np.array([0.20654008, 0.12197225, 0.05136952])
        Izazbz = XYZ_to_Izazbz(XYZ)

        d_r = (("reference", 1), ("1", 1), ("100", 1))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    XYZ_to_Izazbz(XYZ * factor), Izazbz * factor, decimal=7
                )

    @ignore_numpy_errors
    def test_nan_XYZ_to_Izazbz(self):
        """
        Test :func:`colour.models.jzazbz.XYZ_to_Izazbz` definition nan
        support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            XYZ = np.array(case)
            XYZ_to_Izazbz(XYZ)


class TestIzazbz_to_XYZ(unittest.TestCase):
    """
    Define :func:`colour.models.jzazbz.Izazbz_to_XYZ` definition unit tests
    methods.
    """

    def test_Izazbz_to_XYZ(self):
        """Test :func:`colour.models.jzazbz.Izazbz_to_XYZ` definition."""

        np.testing.assert_allclose(
            Izazbz_to_XYZ(np.array([0.01207793, 0.00924302, 0.00526007])),
            np.array([0.20654008, 0.12197225, 0.05136952]),
            rtol=0.000001,
            atol=0.000001,
        )

        np.testing.assert_allclose(
            Izazbz_to_XYZ(np.array([0.01397346, -0.00608426, 0.00534077])),
            np.array([0.14222010, 0.23042768, 0.10495772]),
            rtol=0.000001,
            atol=0.000001,
        )

        np.testing.assert_allclose(
            Izazbz_to_XYZ(np.array([0.03927203, 0.00064174, -0.00052906])),
            np.array([0.96907232, 1.00000000, 1.12179215]),
            rtol=0.000001,
            atol=0.000001,
        )

        np.testing.assert_allclose(
            Izazbz_to_XYZ(np.array([0.03927203, 0.00064174, -0.00052906])),
            np.array([0.96907232, 1.00000000, 1.12179215]),
            rtol=0.000001,
            atol=0.000001,
        )

        np.testing.assert_almost_equal(
            Izazbz_to_XYZ(
                np.array([0.01049146, 0.00924302, 0.00526007]),
                method="Safdar 2021",
            ),
            np.array([0.20654008, 0.12197225, 0.05136952]),
            decimal=7,
        )

        np.testing.assert_array_equal(
            Izazbz_to_XYZ(
                np.array([0.01049146, 0.00924302, 0.00526007]),
                method="Safdar 2021",
            ),
            Izazbz_to_XYZ(
                np.array([0.01049146, 0.00924302, 0.00526007]), method="ZCAM"
            ),
        )

    def test_n_dimensional_Izazbz_to_XYZ(self):
        """
        Test :func:`colour.models.jzazbz.Izazbz_to_XYZ` definition
        n-dimensional support.
        """

        Izazbz = np.array([0.01207793, 0.00924302, 0.00526007])
        XYZ = Izazbz_to_XYZ(Izazbz)

        Izazbz = np.tile(Izazbz, (6, 1))
        XYZ = np.tile(XYZ, (6, 1))
        np.testing.assert_allclose(
            Izazbz_to_XYZ(Izazbz), XYZ, rtol=0.000001, atol=0.000001
        )

        Izazbz = np.reshape(Izazbz, (2, 3, 3))
        XYZ = np.reshape(XYZ, (2, 3, 3))
        np.testing.assert_allclose(
            Izazbz_to_XYZ(Izazbz), XYZ, rtol=0.000001, atol=0.000001
        )

    def test_domain_range_scale_Izazbz_to_XYZ(self):
        """
        Test :func:`colour.models.jzazbz.Izazbz_to_XYZ` definition domain and
        range scale support.
        """

        Izazbz = np.array([0.01207793, 0.00924302, 0.00526007])
        XYZ = Izazbz_to_XYZ(Izazbz)

        d_r = (("reference", 1), ("1", 1), ("100", 1))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_allclose(
                    Izazbz_to_XYZ(Izazbz * factor),
                    XYZ * factor,
                    rtol=0.000001,
                    atol=0.000001,
                )

    @ignore_numpy_errors
    def test_nan_Izazbz_to_XYZ(self):
        """
        Test :func:`colour.models.jzazbz.Izazbz_to_XYZ` definition nan
        support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            Izazbz = np.array(case)
            Izazbz_to_XYZ(Izazbz)


class TestXYZ_to_Jzazbz(unittest.TestCase):
    """
    Define :func:`colour.models.jzazbz.TestXYZ_to_Jzazbz` definition unit
    tests methods.
    """

    def test_XYZ_to_Jzazbz(self):
        """Test :func:`colour.models.jzazbz.XYZ_to_Jzazbz` definition."""

        np.testing.assert_almost_equal(
            XYZ_to_Jzazbz(np.array([0.20654008, 0.12197225, 0.05136952])),
            np.array([0.00535048, 0.00924302, 0.00526007]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            XYZ_to_Jzazbz(np.array([0.14222010, 0.23042768, 0.10495772])),
            np.array([0.00619681, -0.00608426, 0.00534077]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            XYZ_to_Jzazbz(np.array([0.96907232, 1.00000000, 1.12179215])),
            np.array([0.01766826, 0.00064174, -0.00052906]),
            decimal=7,
        )

    def test_n_dimensional_XYZ_to_Jzazbz(self):
        """
        Test :func:`colour.models.jzazbz.XYZ_to_Jzazbz` definition
        n-dimensional support.
        """

        XYZ = np.array([0.20654008, 0.12197225, 0.05136952])
        Jzazbz = XYZ_to_Jzazbz(XYZ)

        XYZ = np.tile(XYZ, (6, 1))
        Jzazbz = np.tile(Jzazbz, (6, 1))
        np.testing.assert_almost_equal(XYZ_to_Jzazbz(XYZ), Jzazbz, decimal=7)

        XYZ = np.reshape(XYZ, (2, 3, 3))
        Jzazbz = np.reshape(Jzazbz, (2, 3, 3))
        np.testing.assert_almost_equal(XYZ_to_Jzazbz(XYZ), Jzazbz, decimal=7)

    def test_domain_range_scale_XYZ_to_Jzazbz(self):
        """
        Test :func:`colour.models.jzazbz.XYZ_to_Jzazbz` definition domain and
        range scale support.
        """

        XYZ = np.array([0.20654008, 0.12197225, 0.05136952])
        Jzazbz = XYZ_to_Jzazbz(XYZ)

        d_r = (("reference", 1), ("1", 1), ("100", 1))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    XYZ_to_Jzazbz(XYZ * factor), Jzazbz * factor, decimal=7
                )

    @ignore_numpy_errors
    def test_nan_XYZ_to_Jzazbz(self):
        """
        Test :func:`colour.models.jzazbz.XYZ_to_Jzazbz` definition nan
        support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            XYZ = np.array(case)
            XYZ_to_Jzazbz(XYZ)


class TestJzazbz_to_XYZ(unittest.TestCase):
    """
    Define :func:`colour.models.jzazbz.Jzazbz_to_XYZ` definition unit tests
    methods.
    """

    def test_Jzazbz_to_XYZ(self):
        """Test :func:`colour.models.jzazbz.Jzazbz_to_XYZ` definition."""

        np.testing.assert_allclose(
            Jzazbz_to_XYZ(np.array([0.00535048, 0.00924302, 0.00526007])),
            np.array([0.20654008, 0.12197225, 0.05136952]),
            rtol=0.000001,
            atol=0.000001,
        )

        np.testing.assert_allclose(
            Jzazbz_to_XYZ(np.array([0.00619681, -0.00608426, 0.00534077])),
            np.array([0.14222010, 0.23042768, 0.10495772]),
            rtol=0.000001,
            atol=0.000001,
        )

        np.testing.assert_allclose(
            Jzazbz_to_XYZ(np.array([0.01766826, 0.00064174, -0.00052906])),
            np.array([0.96907232, 1.00000000, 1.12179215]),
            rtol=0.000001,
            atol=0.000001,
        )

    def test_n_dimensional_Jzazbz_to_XYZ(self):
        """
        Test :func:`colour.models.jzazbz.Jzazbz_to_XYZ` definition
        n-dimensional support.
        """

        Jzazbz = np.array([0.00535048, 0.00924302, 0.00526007])
        XYZ = Jzazbz_to_XYZ(Jzazbz)

        Jzazbz = np.tile(Jzazbz, (6, 1))
        XYZ = np.tile(XYZ, (6, 1))
        np.testing.assert_allclose(
            Jzazbz_to_XYZ(Jzazbz), XYZ, rtol=0.000001, atol=0.000001
        )

        Jzazbz = np.reshape(Jzazbz, (2, 3, 3))
        XYZ = np.reshape(XYZ, (2, 3, 3))
        np.testing.assert_allclose(
            Jzazbz_to_XYZ(Jzazbz), XYZ, rtol=0.000001, atol=0.000001
        )

    def test_domain_range_scale_Jzazbz_to_XYZ(self):
        """
        Test :func:`colour.models.jzazbz.Jzazbz_to_XYZ` definition domain and
        range scale support.
        """

        Jzazbz = np.array([0.00535048, 0.00924302, 0.00526007])
        XYZ = Jzazbz_to_XYZ(Jzazbz)

        d_r = (("reference", 1), ("1", 1), ("100", 1))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_allclose(
                    Jzazbz_to_XYZ(Jzazbz * factor),
                    XYZ * factor,
                    rtol=0.000001,
                    atol=0.000001,
                )

    @ignore_numpy_errors
    def test_nan_Jzazbz_to_XYZ(self):
        """
        Test :func:`colour.models.jzazbz.Jzazbz_to_XYZ` definition nan
        support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            Jzazbz = np.array(case)
            Jzazbz_to_XYZ(Jzazbz)


if __name__ == "__main__":
    unittest.main()



"""
Defines the unit tests for the :mod:`colour.models.rgb.transfer_functions.aces`
module.
"""

import numpy as np
import unittest

from colour.models.rgb.transfer_functions import (
    log_encoding_ACESproxy,
    log_decoding_ACESproxy,
    log_encoding_ACEScc,
    log_decoding_ACEScc,
    log_encoding_ACEScct,
    log_decoding_ACEScct,
)
from colour.utilities import domain_range_scale, ignore_numpy_errors

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestLogEncoding_ACESproxy",
    "TestLogDecoding_ACESproxy",
    "TestLogEncoding_ACEScc",
    "TestLogDecoding_ACEScc",
    "TestLogDecoding_ACEScct",
]


class TestLogEncoding_ACESproxy(unittest.TestCase):
    """
    Define :func:`colour.models.rgb.transfer_functions.aces.\
log_encoding_ACESproxy`
    definition unit tests methods.
    """

    def test_log_encoding_ACESproxy(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.aces.\
log_encoding_ACESproxy` definition.
        """

        self.assertAlmostEqual(
            log_encoding_ACESproxy(0.0), 0.062561094819159, places=7
        )

        self.assertAlmostEqual(
            log_encoding_ACESproxy(0.18), 0.416422287390029, places=7
        )

        self.assertAlmostEqual(
            log_encoding_ACESproxy(0.18, 12), 0.416361416361416, places=7
        )

        self.assertAlmostEqual(
            log_encoding_ACESproxy(1.0), 0.537634408602151, places=7
        )

        self.assertEqual(log_encoding_ACESproxy(0.18, out_int=True), 426)

    def test_n_dimensional_log_encoding_ACESproxy(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.aces.\
log_encoding_ACESproxy` definition n-dimensional arrays support.
        """

        lin_AP1 = 0.18
        ACESproxy = log_encoding_ACESproxy(lin_AP1)

        lin_AP1 = np.tile(lin_AP1, 6)
        ACESproxy = np.tile(ACESproxy, 6)
        np.testing.assert_almost_equal(
            log_encoding_ACESproxy(lin_AP1), ACESproxy, decimal=7
        )

        lin_AP1 = np.reshape(lin_AP1, (2, 3))
        ACESproxy = np.reshape(ACESproxy, (2, 3))
        np.testing.assert_almost_equal(
            log_encoding_ACESproxy(lin_AP1), ACESproxy, decimal=7
        )

        lin_AP1 = np.reshape(lin_AP1, (2, 3, 1))
        ACESproxy = np.reshape(ACESproxy, (2, 3, 1))
        np.testing.assert_almost_equal(
            log_encoding_ACESproxy(lin_AP1), ACESproxy, decimal=7
        )

    def test_domain_range_scale_log_encoding_ACESproxy(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.aces.\
log_encoding_ACESproxy` definition domain and range scale support.
        """

        lin_AP1 = 0.18
        ACESproxy = log_encoding_ACESproxy(lin_AP1)

        d_r = (("reference", 1), ("1", 1), ("100", 100))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    log_encoding_ACESproxy(lin_AP1 * factor),
                    ACESproxy * factor,
                    decimal=7,
                )

    @ignore_numpy_errors
    def test_nan_log_encoding_ACESproxy(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.aces.\
log_encoding_ACESproxy` definition nan support.
        """

        log_encoding_ACESproxy(
            np.array([-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan])
        )


class TestLogDecoding_ACESproxy(unittest.TestCase):
    """
    Define :func:`colour.models.rgb.transfer_functions.aces.\
log_decoding_ACESproxy`
    definition unit tests methods.
    """

    def test_log_decoding_ACESproxy(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.aces.\
log_decoding_ACESproxy` definition.
        """

        np.testing.assert_allclose(
            log_decoding_ACESproxy(0.062561094819159),
            0.0,
            atol=0.01,
            rtol=0.01,
        )

        np.testing.assert_allclose(
            log_decoding_ACESproxy(0.416422287390029),
            0.18,
            atol=0.01,
            rtol=0.01,
        )

        np.testing.assert_allclose(
            log_decoding_ACESproxy(0.416361416361416, 12),
            0.18,
            atol=0.01,
            rtol=0.01,
        )

        np.testing.assert_allclose(
            log_decoding_ACESproxy(0.537634408602151),
            1.0,
            atol=0.01,
            rtol=0.01,
        )

        np.testing.assert_allclose(
            log_decoding_ACESproxy(426, in_int=True),
            0.18,
            atol=0.01,
            rtol=0.01,
        )

    def test_n_dimensional_log_decoding_ACESproxy(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.aces.\
log_decoding_ACESproxy` definition n-dimensional arrays support.
        """

        ACESproxy = 0.416422287390029
        lin_AP1 = log_decoding_ACESproxy(ACESproxy)

        ACESproxy = np.tile(ACESproxy, 6)
        lin_AP1 = np.tile(lin_AP1, 6)
        np.testing.assert_almost_equal(
            log_decoding_ACESproxy(ACESproxy), lin_AP1, decimal=7
        )

        ACESproxy = np.reshape(ACESproxy, (2, 3))
        lin_AP1 = np.reshape(lin_AP1, (2, 3))
        np.testing.assert_almost_equal(
            log_decoding_ACESproxy(ACESproxy), lin_AP1, decimal=7
        )

        ACESproxy = np.reshape(ACESproxy, (2, 3, 1))
        lin_AP1 = np.reshape(lin_AP1, (2, 3, 1))
        np.testing.assert_almost_equal(
            log_decoding_ACESproxy(ACESproxy), lin_AP1, decimal=7
        )

    def test_domain_range_scale_log_decoding_ACESproxy(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.aces.\
log_decoding_ACESproxy` definition domain and range scale support.
        """

        ACESproxy = 426.0
        lin_AP1 = log_decoding_ACESproxy(ACESproxy)

        d_r = (("reference", 1), ("1", 1), ("100", 100))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    log_decoding_ACESproxy(ACESproxy * factor),
                    lin_AP1 * factor,
                    decimal=7,
                )

    @ignore_numpy_errors
    def test_nan_log_decoding_ACESproxy(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.aces.\
log_decoding_ACESproxy` definition nan support.
        """

        log_decoding_ACESproxy(
            np.array([-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan])
        )


class TestLogEncoding_ACEScc(unittest.TestCase):
    """
    Define :func:`colour.models.rgb.transfer_functions.aces.\
log_encoding_ACEScc` definition unit tests methods.
    """

    def test_log_encoding_ACEScc(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.aces.\
log_encoding_ACEScc` definition.
        """

        self.assertAlmostEqual(
            log_encoding_ACEScc(0.0), -0.358447488584475, places=7
        )

        self.assertAlmostEqual(
            log_encoding_ACEScc(0.18), 0.413588402492442, places=7
        )

        self.assertAlmostEqual(
            log_encoding_ACEScc(1.0), 0.554794520547945, places=7
        )

    def test_n_dimensional_log_encoding_ACEScc(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.aces.\
log_encoding_ACEScc` definition n-dimensional arrays support.
        """

        lin_AP1 = 0.18
        ACEScc = log_encoding_ACEScc(lin_AP1)

        lin_AP1 = np.tile(lin_AP1, 6)
        ACEScc = np.tile(ACEScc, 6)
        np.testing.assert_almost_equal(
            log_encoding_ACEScc(lin_AP1), ACEScc, decimal=7
        )

        lin_AP1 = np.reshape(lin_AP1, (2, 3))
        ACEScc = np.reshape(ACEScc, (2, 3))
        np.testing.assert_almost_equal(
            log_encoding_ACEScc(lin_AP1), ACEScc, decimal=7
        )

        lin_AP1 = np.reshape(lin_AP1, (2, 3, 1))
        ACEScc = np.reshape(ACEScc, (2, 3, 1))
        np.testing.assert_almost_equal(
            log_encoding_ACEScc(lin_AP1), ACEScc, decimal=7
        )

    def test_domain_range_scale_log_encoding_ACEScc(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.aces.\
log_encoding_ACEScc` definition domain and range scale support.
        """

        lin_AP1 = 0.18
        ACEScc = log_encoding_ACEScc(lin_AP1)

        d_r = (("reference", 1), ("1", 1), ("100", 100))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    log_encoding_ACEScc(lin_AP1 * factor),
                    ACEScc * factor,
                    decimal=7,
                )

    @ignore_numpy_errors
    def test_nan_log_encoding_ACEScc(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.aces.\
log_encoding_ACEScc` definition nan support.
        """

        log_encoding_ACEScc(
            np.array([-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan])
        )


class TestLogDecoding_ACEScc(unittest.TestCase):
    """
    Define :func:`colour.models.rgb.transfer_functions.aces.\
log_decoding_ACEScc` definition unit tests methods.
    """

    def test_log_decoding_ACEScc(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.aces.\
log_decoding_ACEScc` definition.
        """

        self.assertAlmostEqual(
            log_decoding_ACEScc(-0.358447488584475), 0.0, places=7
        )

        self.assertAlmostEqual(
            log_decoding_ACEScc(0.413588402492442), 0.18, places=7
        )

        self.assertAlmostEqual(
            log_decoding_ACEScc(0.554794520547945), 1.0, places=7
        )

    def test_n_dimensional_log_decoding_ACEScc(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.aces.\
log_decoding_ACEScc` definition n-dimensional arrays support.
        """

        ACEScc = 0.413588402492442
        lin_AP1 = log_decoding_ACEScc(ACEScc)

        ACEScc = np.tile(ACEScc, 6)
        lin_AP1 = np.tile(lin_AP1, 6)
        np.testing.assert_almost_equal(
            log_decoding_ACEScc(ACEScc), lin_AP1, decimal=7
        )

        ACEScc = np.reshape(ACEScc, (2, 3))
        lin_AP1 = np.reshape(lin_AP1, (2, 3))
        np.testing.assert_almost_equal(
            log_decoding_ACEScc(ACEScc), lin_AP1, decimal=7
        )

        ACEScc = np.reshape(ACEScc, (2, 3, 1))
        lin_AP1 = np.reshape(lin_AP1, (2, 3, 1))
        np.testing.assert_almost_equal(
            log_decoding_ACEScc(ACEScc), lin_AP1, decimal=7
        )

    def test_domain_range_scale_log_decoding_ACEScc(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.aces.\
log_decoding_ACEScc` definition domain and range scale support.
        """

        ACEScc = 0.413588402492442
        lin_AP1 = log_decoding_ACEScc(ACEScc)

        d_r = (("reference", 1), ("1", 1), ("100", 100))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    log_decoding_ACEScc(ACEScc * factor),
                    lin_AP1 * factor,
                    decimal=7,
                )

    @ignore_numpy_errors
    def test_nan_log_decoding_ACEScc(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.aces.\
log_decoding_ACEScc` definition nan support.
        """

        log_decoding_ACEScc(
            np.array([-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan])
        )


class TestLogEncoding_ACEScct(unittest.TestCase):
    """
    Define :func:`colour.models.rgb.transfer_functions.aces.\
log_encoding_ACEScct` definition unit tests methods.
    """

    def test_log_encoding_ACEScct(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.aces.\
log_encoding_ACEScct` definition.
        """

        self.assertAlmostEqual(
            log_encoding_ACEScct(0.0), 0.072905534195835495, places=7
        )

        self.assertAlmostEqual(
            log_encoding_ACEScct(0.18), 0.413588402492442, places=7
        )

        self.assertAlmostEqual(
            log_encoding_ACEScct(1.0), 0.554794520547945, places=7
        )

    def test_n_dimensional_log_encoding_ACEScct(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.aces.\
log_encoding_ACEScct` definition n-dimensional arrays support.
        """

        lin_AP1 = 0.18
        ACEScct = log_encoding_ACEScct(lin_AP1)

        lin_AP1 = np.tile(lin_AP1, 6)
        ACEScct = np.tile(ACEScct, 6)
        np.testing.assert_almost_equal(
            log_encoding_ACEScct(lin_AP1), ACEScct, decimal=7
        )

        lin_AP1 = np.reshape(lin_AP1, (2, 3))
        ACEScct = np.reshape(ACEScct, (2, 3))
        np.testing.assert_almost_equal(
            log_encoding_ACEScct(lin_AP1), ACEScct, decimal=7
        )

        lin_AP1 = np.reshape(lin_AP1, (2, 3, 1))
        ACEScct = np.reshape(ACEScct, (2, 3, 1))
        np.testing.assert_almost_equal(
            log_encoding_ACEScct(lin_AP1), ACEScct, decimal=7
        )

    def test_domain_range_scale_log_encoding_ACEScct(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.aces.\
log_encoding_ACEScct` definition domain and range scale support.
        """

        lin_AP1 = 0.18
        ACEScct = log_encoding_ACEScct(lin_AP1)

        d_r = (("reference", 1), ("1", 1), ("100", 100))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    log_encoding_ACEScct(lin_AP1 * factor),
                    ACEScct * factor,
                    decimal=7,
                )

    def test_ACEScc_equivalency_log_encoding_ACEScct(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.aces.\
log_encoding_ACEScct` definition ACEScc equivalency, and explicit requirement
        specified by AMPAS ACES specification S-2016-001
        (https://github.com/ampas/aces-dev/blob/v1.0.3/documents/LaTeX/\
S-2016-001/introduction.tex#L14)
        """

        equiv = np.linspace(0.0078125, 222.86094420380761, 100)
        np.testing.assert_almost_equal(
            log_encoding_ACEScct(equiv), log_encoding_ACEScc(equiv), decimal=7
        )

    @ignore_numpy_errors
    def test_nan_log_encoding_ACEScct(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.aces.\
log_encoding_ACEScct` definition nan support.
        """

        log_encoding_ACEScct(
            np.array([-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan])
        )


class TestLogDecoding_ACEScct(unittest.TestCase):
    """
    Define :func:`colour.models.rgb.transfer_functions.aces.\
log_decoding_ACEScct` definition unit tests methods.
    """

    def test_log_decoding_ACEScct(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.aces.\
log_decoding_ACEScct` definition.
        """

        self.assertAlmostEqual(
            log_decoding_ACEScct(0.072905534195835495), 0.0, places=7
        )

        self.assertAlmostEqual(
            log_decoding_ACEScct(0.41358840249244228), 0.18, places=7
        )

        self.assertAlmostEqual(
            log_decoding_ACEScct(0.554794520547945), 1.0, places=7
        )

    def test_n_dimensional_log_decoding_ACEScct(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.aces.\
log_decoding_ACEScct` definition n-dimensional arrays support.
        """

        ACEScct = 0.413588402492442
        lin_AP1 = log_decoding_ACEScct(ACEScct)

        ACEScct = np.tile(ACEScct, 6)
        lin_AP1 = np.tile(lin_AP1, 6)
        np.testing.assert_almost_equal(
            log_decoding_ACEScct(ACEScct), lin_AP1, decimal=7
        )

        ACEScct = np.reshape(ACEScct, (2, 3))
        lin_AP1 = np.reshape(lin_AP1, (2, 3))
        np.testing.assert_almost_equal(
            log_decoding_ACEScct(ACEScct), lin_AP1, decimal=7
        )

        ACEScct = np.reshape(ACEScct, (2, 3, 1))
        lin_AP1 = np.reshape(lin_AP1, (2, 3, 1))
        np.testing.assert_almost_equal(
            log_decoding_ACEScct(ACEScct), lin_AP1, decimal=7
        )

    def test_domain_range_scale_log_decoding_ACEScct(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.aces.\
log_decoding_ACEScct` definition domain and range scale support.
        """

        ACEScc = 0.413588402492442
        lin_AP1 = log_decoding_ACEScct(ACEScc)

        d_r = (("reference", 1), ("1", 1), ("100", 100))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    log_decoding_ACEScct(ACEScc * factor),
                    lin_AP1 * factor,
                    decimal=7,
                )

    def test_ACEScc_equivalency_log_decoding_ACEScct(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.aces.\
log_decoding_ACEScct` definition ACEScc equivalency, and explicit requirement
        specified by AMPAS ACES specification S-2016-001
        (https://github.com/ampas/aces-dev/blob/v1.0.3/documents/LaTeX/\
S-2016-001/introduction.tex#L14)
        """

        equiv = np.linspace(0.15525114155251146, 1.0, 100)
        np.testing.assert_almost_equal(
            log_decoding_ACEScct(equiv), log_decoding_ACEScc(equiv), decimal=7
        )

    @ignore_numpy_errors
    def test_nan_log_decoding_ACEScct(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.aces.\
log_decoding_ACEScct` definition nan support.
        """

        log_decoding_ACEScct(
            np.array([-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan])
        )


if __name__ == "__main__":
    unittest.main()



"""
Deprecation Utilities
=====================

Defines various deprecation management related objects.
"""

from __future__ import annotations

import sys
from importlib import import_module
from collections import namedtuple
from operator import attrgetter

from colour.utilities import attest, optional, usage_warning
from colour.hints import Any, Dict, List, ModuleType, Optional

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "ObjectRenamed",
    "ObjectRemoved",
    "ObjectFutureRename",
    "ObjectFutureRemove",
    "ObjectFutureAccessChange",
    "ObjectFutureAccessRemove",
    "ModuleAPI",
    "ArgumentRenamed",
    "ArgumentRemoved",
    "ArgumentFutureRename",
    "ArgumentFutureRemove",
    "get_attribute",
    "build_API_changes",
    "handle_arguments_deprecation",
]


class ObjectRenamed(namedtuple("ObjectRenamed", ("name", "new_name"))):
    """
    A class used for an object that has been renamed.

    Parameters
    ----------
    name
        Object name that changed.
    new_name
        Object new name.
    """

    def __str__(self) -> str:
        """
        Return a formatted string representation of the class.

        Returns
        -------
        :class:`str`
            Formatted string representation.
        """

        return f'"{self.name}" object has been renamed to "{self.new_name}".'


class ObjectRemoved(namedtuple("ObjectRemoved", ("name",))):
    """
    A class used for an object that has been removed.

    Parameters
    ----------
    name
        Object name that has been removed.
    """

    def __str__(self) -> str:
        """
        Return a formatted string representation of the class.

        Returns
        -------
        :class:`str`
            Formatted string representation.
        """

        return f'"{self.name}" object has been removed from the API.'


class ObjectFutureRename(
    namedtuple("ObjectFutureRename", ("name", "new_name"))
):
    """
    A class used for future object name deprecation, i.e. object name will
    change in a future release.

    Parameters
    ----------
    name
        Object name that will change in a future release.
    new_name
        Object future release name.
    """

    def __str__(self) -> str:
        """
        Return a formatted string representation of the deprecation type.

        Returns
        -------
        :class:`str`
            Formatted string representation.
        """

        return (
            f'"{self.name}" object is deprecated and will be renamed to '
            f'"{self.new_name}" in a future release.'
        )


class ObjectFutureRemove(namedtuple("ObjectFutureRemove", ("name",))):
    """
    A class used for future object removal.

    Parameters
    ----------
    name
        Object name that will be removed in a future release.
    """

    def __str__(self) -> str:
        """
        Return a formatted string representation of the deprecation type.

        Returns
        -------
        :class:`str`
            Formatted string representation.
        """

        return (
            f'"{self.name}" object is deprecated and will be removed in '
            f"a future release."
        )


class ObjectFutureAccessChange(
    namedtuple("ObjectFutureAccessChange", ("access", "new_access"))
):
    """
    A class used for future object access deprecation, i.e. object access will
    change in a future release.

    Parameters
    ----------
    access
        Object access that will change in a future release.
    new_access
        Object future release access.
    """

    def __str__(self) -> str:
        """
        Return a formatted string representation of the deprecation type.

        Returns
        -------
        :class:`str`
            Formatted string representation.
        """

        return (
            f'"{self.access}" object access is deprecated and will change '
            f'to "{self.new_access}" in a future release.'
        )


class ObjectFutureAccessRemove(
    namedtuple("ObjectFutureAccessRemove", ("name",))
):
    """
    A class used for future object access removal, i.e. object access will
    be removed in a future release.

    Parameters
    ----------
    name
        Object name whose access will removed in a future release.
    """

    def __str__(self) -> str:
        """
        Return a formatted string representation of the deprecation type.

        Returns
        -------
        :class:`str`
            Formatted string representation.
        """

        return (
            f'"{self.name}" object access will be removed in a future release.'
        )


class ArgumentRenamed(namedtuple("ArgumentRenamed", ("name", "new_name"))):
    """
    A class used for an argument that has been renamed.

    Parameters
    ----------
    name
        Argument name that changed.
    new_name
        Argument new name.
    """

    def __str__(self) -> str:
        """
        Return a formatted string representation of the class.

        Returns
        -------
        :class:`str`
            Formatted string representation.
        """

        return f'"{self.name}" argument has been renamed to "{self.new_name}".'


class ArgumentRemoved(namedtuple("ArgumentRemoved", ("name",))):
    """
    A class used for an argument that has been removed.

    Parameters
    ----------
    name
        Argument name that has been removed.
    """

    def __str__(self) -> str:
        """
        Return a formatted string representation of the class.

        Returns
        -------
        :class:`str`
            Formatted string representation.
        """

        return f'"{self.name}" argument has been removed from the API.'


class ArgumentFutureRename(
    namedtuple("ArgumentFutureRename", ("name", "new_name"))
):
    """
    A class used for future argument name deprecation, i.e. argument name will
    change in a future release.

    Parameters
    ----------
    name
        Argument name that will change in a future release.
    new_name
        Argument future release name.
    """

    def __str__(self) -> str:
        """
        Return a formatted string representation of the deprecation type.

        Returns
        -------
        :class:`str`
            Formatted string representation.
        """

        return (
            f'"{self.name}" argument is deprecated and will be renamed to '
            f'"{self.new_name}" in a future release.'
        )


class ArgumentFutureRemove(namedtuple("ArgumentFutureRemove", ("name",))):
    """
    A class used for future argument removal.

    Parameters
    ----------
    name
        Argument name that will be removed in a future release.
    """

    def __str__(self) -> str:
        """
        Return a formatted string representation of the deprecation type.

        Returns
        -------
        :class:`str`
            Formatted string representation.
        """

        return (
            f'"{self.name}" argument is deprecated and will be removed in '
            f"a future release."
        )


class ModuleAPI:
    """
    Define a class that allows customisation of module attributes access with
    deprecation management.

    Parameters
    ----------
    module
        Module to customise attributes access.

    Methods
    -------
    -   :meth:`~colour.utilities.ModuleAPI.__init__`
    -   :meth:`~colour.utilities.ModuleAPI.__getattr__`
    -   :meth:`~colour.utilities.ModuleAPI.__dir__`

    Examples
    --------
    >>> import sys
    >>> sys.modules['colour'] = ModuleAPI(sys.modules['colour'])
    ... # doctest: +SKIP
    """

    def __init__(self, module: ModuleType, changes: Optional[Dict] = None):
        self._module = module
        self._changes = optional(changes, {})

    def __getattr__(self, attribute: str) -> Any:
        """
        Return given attribute value while handling deprecation.

        Parameters
        ----------
        attribute
            Attribute name.

        Returns
        -------
        :class:`object`
            Attribute value.

        Raises
        ------
        AttributeError
            If the attribute is not defined.
        """

        change = self._changes.get(attribute)

        if change is not None:
            if not isinstance(change, ObjectRemoved):

                usage_warning(str(change))

                return (
                    getattr(self._module, attribute)
                    if isinstance(change, ObjectFutureRemove)
                    else get_attribute(change[1])
                )
            else:
                raise AttributeError(str(change))

        return getattr(self._module, attribute)

    def __dir__(self) -> List:
        """
        Return the list of names in the module local scope filtered according
        to the changes.

        Returns
        -------
        :class:`list`
            Filtered list of names in the module local scope.
        """

        attributes = [
            attribute
            for attribute in dir(self._module)
            if attribute not in self._changes
        ]

        return attributes


def get_attribute(attribute: str) -> Any:
    """
    Return given attribute value.

    Parameters
    ----------
    attribute
        Attribute to retrieve, ``attribute`` must have a namespace module, e.g.
        *colour.models.eotf_BT2020*.

    Returns
    -------
    :class:`object`
        Retrieved attribute value.

    Examples
    --------
    >>> get_attribute('colour.models.eotf_BT2020')  # doctest: +ELLIPSIS
    <function eotf_BT2020 at 0x...>
    """

    attest("." in attribute, '"{0}" attribute has no namespace!')

    module_name, attribute = attribute.rsplit(".", 1)

    module = optional(sys.modules.get(module_name), import_module(module_name))

    attest(
        module is not None,
        f'"{module_name}" module does not exists or cannot be imported!',
    )

    return attrgetter(attribute)(module)


def build_API_changes(changes: dict) -> Dict:
    """
    Build the effective API changes for a desired API changes mapping.

    Parameters
    ----------
    changes
        Dictionary of desired API changes.

    Returns
    -------
    :class:`dict`
        API changes

    Examples
    --------
    >>> from pprint import pprint
    >>> changes = {
    ...     'ObjectRenamed': [[
    ...         'module.object_1_name',
    ...         'module.object_1_new_name',
    ...     ]],
    ...     'ObjectFutureRename': [[
    ...         'module.object_2_name',
    ...         'module.object_2_new_name',
    ...     ]],
    ...     'ObjectFutureAccessChange': [[
    ...         'module.object_3_access',
    ...         'module.sub_module.object_3_new_access',
    ...     ]],
    ...     'ObjectRemoved': ['module.object_4_name'],
    ...     'ObjectFutureRemove': ['module.object_5_name'],
    ...     'ObjectFutureAccessRemove': ['module.object_6_access'],
    ... }
    >>> pprint(build_API_changes(changes))  # doctest: +SKIP
    {'object_1_name': ObjectRenamed(name='module.object_1_name', \
new_name='module.object_1_new_name'),
     'object_2_name': ObjectFutureRename(name='module.object_2_name', \
new_name='module.object_2_new_name'),
     'object_3_access': ObjectFutureAccessChange(\
access='module.object_3_access', \
new_access='module.sub_module.object_3_new_access'),
     'object_4_name': ObjectRemoved(name='module.object_4_name'),
     'object_5_name': ObjectFutureRemove(name='module.object_5_name'),
     'object_6_access': ObjectFutureAccessRemove(\
name='module.object_6_access')}
    """

    for rename_type in (
        ObjectRenamed,
        ObjectFutureRename,
        ObjectFutureAccessChange,
        ArgumentRenamed,
        ArgumentFutureRename,
    ):
        for change in changes.pop(rename_type.__name__, []):
            changes[change[0].split(".")[-1]] = rename_type(*change)  # noqa

    for remove_type in (
        ObjectRemoved,
        ObjectFutureRemove,
        ObjectFutureAccessRemove,
        ArgumentRemoved,
        ArgumentFutureRemove,
    ):
        for change in changes.pop(remove_type.__name__, []):
            changes[change.split(".")[-1]] = remove_type(change)  # noqa

    return changes


def handle_arguments_deprecation(changes: dict, **kwargs: Any) -> Dict:
    """
    Handle arguments deprecation according to desired API changes mapping.

    Parameters
    ----------
    changes
        Dictionary of desired API changes.

    Other Parameters
    ----------------
    kwargs
        Keywords arguments to handle.

    Returns
    -------
    :class:`dict`
        Handled keywords arguments.

    Examples
    --------
    >>> changes = {
    ...     'ArgumentRenamed': [[
    ...         'argument_1_name',
    ...         'argument_1_new_name',
    ...     ]],
    ...     'ArgumentFutureRename': [[
    ...         'argument_2_name',
    ...         'argument_2_new_name',
    ...     ]],
    ...     'ArgumentRemoved': ['argument_3_name'],
    ...     'ArgumentFutureRemove': ['argument_4_name'],
    ... }
    >>> handle_arguments_deprecation(changes, argument_1_name=True,
    ...                             argument_2_name=True, argument_4_name=True)
    ... # doctest: +SKIP
    {'argument_4_name': True, 'argument_1_new_name': True, \
'argument_2_new_name': True}
    """

    changes = build_API_changes(changes)

    for kwarg in kwargs.copy():
        change = changes.get(kwarg)

        if change is None:
            continue

        if not isinstance(change, ArgumentRemoved):

            usage_warning(str(change))

            if isinstance(change, ArgumentFutureRemove):
                continue
            else:
                kwargs[change[1]] = kwargs.pop(kwarg)
        else:
            kwargs.pop(kwarg)
            usage_warning(str(change))

    return kwargs



"""
Geometry Primitives
===================

Defines various geometry primitives and their generation methods:

-   :attr:`colour.geometry.MAPPING_PLANE_TO_AXIS`
-   :func:`colour.geometry.primitive_grid`
-   :func:`colour.geometry.primitive_cube`
-   :func:`colour.PRIMITIVE_METHODS`
-   :func:`colour.primitive`

References
----------
-   :cite:`Cabello2015` : Cabello, R. (n.d.). PlaneGeometry.js. Retrieved May
    12, 2015, from
    https://github.com/mrdoob/three.js/blob/dev/src/geometries/PlaneGeometry.js
"""

from __future__ import annotations

import numpy as np

from colour.constants import DEFAULT_INT_DTYPE, DEFAULT_FLOAT_DTYPE
from colour.hints import (
    Any,
    DTypeFloating,
    DTypeInteger,
    Floating,
    Integer,
    Literal,
    NDArray,
    Optional,
    Tuple,
    Type,
    Union,
    cast,
)
from colour.utilities import (
    CaseInsensitiveMapping,
    as_int_array,
    filter_kwargs,
    ones,
    optional,
    zeros,
    validate_method,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "MAPPING_PLANE_TO_AXIS",
    "primitive_grid",
    "primitive_cube",
    "PRIMITIVE_METHODS",
    "primitive",
]

MAPPING_PLANE_TO_AXIS: CaseInsensitiveMapping = CaseInsensitiveMapping(
    {
        "yz": "+x",
        "zy": "-x",
        "xz": "+y",
        "zx": "-y",
        "xy": "+z",
        "yx": "-z",
    }
)
MAPPING_PLANE_TO_AXIS.__doc__ = """Plane to axis mapping."""


def primitive_grid(
    width: Floating = 1,
    height: Floating = 1,
    width_segments: Integer = 1,
    height_segments: Integer = 1,
    axis: Literal[
        "-x", "+x", "-y", "+y", "-z", "+z", "xy", "xz", "yz", "yx", "zx", "zy"
    ] = "+z",
    dtype_vertices: Optional[Type[DTypeFloating]] = None,
    dtype_indexes: Optional[Type[DTypeInteger]] = None,
) -> Tuple[NDArray, NDArray, NDArray]:
    """
    Generate vertices and indexes for a filled and outlined grid primitive.

    Parameters
    ----------
    width
        Grid width.
    height
        Grid height.
    width_segments
        Grid segments count along the width.
    height_segments
        Grid segments count along the height.
    axis
        Axis the primitive will be normal to, or plane the primitive will be
        co-planar with.
    dtype_vertices
        :class:`numpy.dtype` to use for the grid vertices, default to
        the :class:`numpy.dtype` defined by the
        :attr:`colour.constant.DEFAULT_FLOAT_DTYPE` attribute.
    dtype_indexes
        :class:`numpy.dtype` to use for the grid indexes, default to
        the :class:`numpy.dtype` defined by the
        :attr:`colour.constant.DEFAULT_INT_DTYPE` attribute.

    Returns
    -------
    :class:`tuple`
        Tuple of grid vertices, face indexes to produce a filled grid and
        outline indexes to produce an outline of the faces of the grid.

    References
    ----------
    :cite:`Cabello2015`

    Examples
    --------
    >>> vertices, faces, outline = primitive_grid()
    >>> print(vertices)
    [([-0.5,  0.5,  0. ], [ 0.,  1.], [ 0.,  0.,  1.], [ 0.,  1.,  0.,  1.])
     ([ 0.5,  0.5,  0. ], [ 1.,  1.], [ 0.,  0.,  1.], [ 1.,  1.,  0.,  1.])
     ([-0.5, -0.5,  0. ], [ 0.,  0.], [ 0.,  0.,  1.], [ 0.,  0.,  0.,  1.])
     ([ 0.5, -0.5,  0. ], [ 1.,  0.], [ 0.,  0.,  1.], [ 1.,  0.,  0.,  1.])]
    >>> print(faces)
    [[0 2 1]
     [2 3 1]]
    >>> print(outline)
    [[0 2]
     [2 3]
     [3 1]
     [1 0]]
    """

    axis = MAPPING_PLANE_TO_AXIS.get(axis, axis).lower()

    dtype_vertices = cast(
        Type[DTypeFloating], optional(dtype_vertices, DEFAULT_FLOAT_DTYPE)
    )
    dtype_indexes = cast(
        Type[DTypeInteger], optional(dtype_indexes, DEFAULT_INT_DTYPE)
    )

    x_grid = width_segments
    y_grid = height_segments

    x_grid1 = int(x_grid + 1)
    y_grid1 = int(y_grid + 1)

    # Positions, normals and uvs.
    positions = zeros(x_grid1 * y_grid1 * 3)
    normals = zeros(x_grid1 * y_grid1 * 3)
    uvs = zeros(x_grid1 * y_grid1 * 2)

    y = np.arange(y_grid1) * height / y_grid - height / 2
    x = np.arange(x_grid1) * width / x_grid - width / 2

    positions[::3] = np.tile(x, y_grid1)
    positions[1::3] = -np.repeat(y, x_grid1)

    normals[2::3] = 1

    uvs[::2] = np.tile(np.arange(x_grid1) / x_grid, y_grid1)
    uvs[1::2] = np.repeat(1 - np.arange(y_grid1) / y_grid, x_grid1)

    # Faces and outline.
    faces_indexes = []
    outline_indexes = []
    for i_y in range(y_grid):
        for i_x in range(x_grid):
            a = i_x + x_grid1 * i_y
            b = i_x + x_grid1 * (i_y + 1)
            c = (i_x + 1) + x_grid1 * (i_y + 1)
            d = (i_x + 1) + x_grid1 * i_y

            faces_indexes.extend([(a, b, d), (b, c, d)])
            outline_indexes.extend([(a, b), (b, c), (c, d), (d, a)])

    faces = np.reshape(as_int_array(faces_indexes, dtype_indexes), (-1, 3))
    outline = np.reshape(as_int_array(outline_indexes, dtype_indexes), (-1, 2))

    positions = np.reshape(positions, (-1, 3))
    uvs = np.reshape(uvs, (-1, 2))
    normals = np.reshape(normals, (-1, 3))

    if axis in ("-x", "+x"):
        shift, zero_axis = 1, 0
    elif axis in ("-y", "+y"):
        shift, zero_axis = -1, 1
    elif axis in ("-z", "+z"):
        shift, zero_axis = 0, 2

    sign = -1 if "-" in axis else 1

    positions = np.roll(positions, shift, -1)
    normals = np.roll(normals, shift, -1) * sign
    vertex_colours = np.ravel(positions)
    vertex_colours = np.hstack(
        [
            np.reshape(
                np.interp(
                    vertex_colours,
                    (np.min(vertex_colours), np.max(vertex_colours)),
                    (0, 1),
                ),
                positions.shape,
            ),
            ones((positions.shape[0], 1)),
        ]
    )
    vertex_colours[..., zero_axis] = 0

    vertices = zeros(
        positions.shape[0],
        [
            ("position", dtype_vertices, 3),
            ("uv", dtype_vertices, 2),
            ("normal", dtype_vertices, 3),
            ("colour", dtype_vertices, 4),
        ],  # type: ignore[arg-type]
    )

    vertices["position"] = positions
    vertices["uv"] = uvs
    vertices["normal"] = normals
    vertices["colour"] = vertex_colours

    return vertices, faces, outline


def primitive_cube(
    width: Floating = 1,
    height: Floating = 1,
    depth: Floating = 1,
    width_segments: Integer = 1,
    height_segments: Integer = 1,
    depth_segments: Integer = 1,
    planes: Optional[
        Literal[
            "-x",
            "+x",
            "-y",
            "+y",
            "-z",
            "+z",
            "xy",
            "xz",
            "yz",
            "yx",
            "zx",
            "zy",
        ]
    ] = None,
    dtype_vertices: Optional[Type[DTypeFloating]] = None,
    dtype_indexes: Optional[Type[DTypeInteger]] = None,
) -> Tuple[NDArray, NDArray, NDArray]:
    """
    Generate vertices and indexes for a filled and outlined cube primitive.

    Parameters
    ----------
    width
        Cube width.
    height
        Cube height.
    depth
        Cube depth.
    width_segments
        Cube segments count along the width.
    height_segments
        Cube segments count along the height.
    depth_segments
        Cube segments count along the depth.
    planes
        Grid primitives to include in the cube construction.
    dtype_vertices
        :class:`numpy.dtype` to use for the grid vertices, default to
        the :class:`numpy.dtype` defined by the
        :attr:`colour.constant.DEFAULT_FLOAT_DTYPE` attribute.
    dtype_indexes
        :class:`numpy.dtype` to use for the grid indexes, default to
        the :class:`numpy.dtype` defined by the
        :attr:`colour.constant.DEFAULT_INT_DTYPE` attribute.

    Returns
    -------
    :class:`tuple`
        Tuple of cube vertices, face indexes to produce a filled cube and
        outline indexes to produce an outline of the faces of the cube.

    Examples
    --------
    >>> vertices, faces, outline = primitive_cube()
    >>> print(vertices)
    [([-0.5,  0.5, -0.5], [ 0.,  1.], [-0., -0., -1.], [ 0.,  1.,  0.,  1.])
     ([ 0.5,  0.5, -0.5], [ 1.,  1.], [-0., -0., -1.], [ 1.,  1.,  0.,  1.])
     ([-0.5, -0.5, -0.5], [ 0.,  0.], [-0., -0., -1.], [ 0.,  0.,  0.,  1.])
     ([ 0.5, -0.5, -0.5], [ 1.,  0.], [-0., -0., -1.], [ 1.,  0.,  0.,  1.])
     ([-0.5,  0.5,  0.5], [ 0.,  1.], [ 0.,  0.,  1.], [ 0.,  1.,  1.,  1.])
     ([ 0.5,  0.5,  0.5], [ 1.,  1.], [ 0.,  0.,  1.], [ 1.,  1.,  1.,  1.])
     ([-0.5, -0.5,  0.5], [ 0.,  0.], [ 0.,  0.,  1.], [ 0.,  0.,  1.,  1.])
     ([ 0.5, -0.5,  0.5], [ 1.,  0.], [ 0.,  0.,  1.], [ 1.,  0.,  1.,  1.])
     ([ 0.5, -0.5, -0.5], [ 0.,  1.], [-0., -1., -0.], [ 1.,  0.,  0.,  1.])
     ([ 0.5, -0.5,  0.5], [ 1.,  1.], [-0., -1., -0.], [ 1.,  0.,  1.,  1.])
     ([-0.5, -0.5, -0.5], [ 0.,  0.], [-0., -1., -0.], [ 0.,  0.,  0.,  1.])
     ([-0.5, -0.5,  0.5], [ 1.,  0.], [-0., -1., -0.], [ 0.,  0.,  1.,  1.])
     ([ 0.5,  0.5, -0.5], [ 0.,  1.], [ 0.,  1.,  0.], [ 1.,  1.,  0.,  1.])
     ([ 0.5,  0.5,  0.5], [ 1.,  1.], [ 0.,  1.,  0.], [ 1.,  1.,  1.,  1.])
     ([-0.5,  0.5, -0.5], [ 0.,  0.], [ 0.,  1.,  0.], [ 0.,  1.,  0.,  1.])
     ([-0.5,  0.5,  0.5], [ 1.,  0.], [ 0.,  1.,  0.], [ 0.,  1.,  1.,  1.])
     ([-0.5, -0.5,  0.5], [ 0.,  1.], [-1., -0., -0.], [ 0.,  0.,  1.,  1.])
     ([-0.5,  0.5,  0.5], [ 1.,  1.], [-1., -0., -0.], [ 0.,  1.,  1.,  1.])
     ([-0.5, -0.5, -0.5], [ 0.,  0.], [-1., -0., -0.], [ 0.,  0.,  0.,  1.])
     ([-0.5,  0.5, -0.5], [ 1.,  0.], [-1., -0., -0.], [ 0.,  1.,  0.,  1.])
     ([ 0.5, -0.5,  0.5], [ 0.,  1.], [ 1.,  0.,  0.], [ 1.,  0.,  1.,  1.])
     ([ 0.5,  0.5,  0.5], [ 1.,  1.], [ 1.,  0.,  0.], [ 1.,  1.,  1.,  1.])
     ([ 0.5, -0.5, -0.5], [ 0.,  0.], [ 1.,  0.,  0.], [ 1.,  0.,  0.,  1.])
     ([ 0.5,  0.5, -0.5], [ 1.,  0.], [ 1.,  0.,  0.], [ 1.,  1.,  0.,  1.])]
    >>> print(faces)
    [[ 1  2  0]
     [ 1  3  2]
     [ 4  6  5]
     [ 6  7  5]
     [ 9 10  8]
     [ 9 11 10]
     [12 14 13]
     [14 15 13]
     [17 18 16]
     [17 19 18]
     [20 22 21]
     [22 23 21]]
    >>> print(outline)
    [[ 0  2]
     [ 2  3]
     [ 3  1]
     [ 1  0]
     [ 4  6]
     [ 6  7]
     [ 7  5]
     [ 5  4]
     [ 8 10]
     [10 11]
     [11  9]
     [ 9  8]
     [12 14]
     [14 15]
     [15 13]
     [13 12]
     [16 18]
     [18 19]
     [19 17]
     [17 16]
     [20 22]
     [22 23]
     [23 21]
     [21 20]]
    """

    axis = (
        sorted(list(MAPPING_PLANE_TO_AXIS.values()))
        if planes is None
        else [
            MAPPING_PLANE_TO_AXIS.get(plane, plane).lower() for plane in planes
        ]
    )

    dtype_vertices = cast(
        Type[DTypeFloating], optional(dtype_vertices, DEFAULT_FLOAT_DTYPE)
    )
    dtype_indexes = cast(
        Type[DTypeInteger], optional(dtype_indexes, DEFAULT_INT_DTYPE)
    )

    w_s, h_s, d_s = width_segments, height_segments, depth_segments

    planes_p = []
    if "-z" in axis:
        planes_p.append(list(primitive_grid(width, depth, w_s, d_s, "-z")))
        planes_p[-1][0]["position"][..., 2] -= height / 2
        planes_p[-1][1] = np.fliplr(planes_p[-1][1])
    if "+z" in axis:
        planes_p.append(list(primitive_grid(width, depth, w_s, d_s, "+z")))
        planes_p[-1][0]["position"][..., 2] += height / 2

    if "-y" in axis:
        planes_p.append(list(primitive_grid(height, width, h_s, w_s, "-y")))
        planes_p[-1][0]["position"][..., 1] -= depth / 2
        planes_p[-1][1] = np.fliplr(planes_p[-1][1])
    if "+y" in axis:
        planes_p.append(list(primitive_grid(height, width, h_s, w_s, "+y")))
        planes_p[-1][0]["position"][..., 1] += depth / 2

    if "-x" in axis:
        planes_p.append(list(primitive_grid(depth, height, d_s, h_s, "-x")))
        planes_p[-1][0]["position"][..., 0] -= width / 2
        planes_p[-1][1] = np.fliplr(planes_p[-1][1])
    if "+x" in axis:
        planes_p.append(list(primitive_grid(depth, height, d_s, h_s, "+x")))
        planes_p[-1][0]["position"][..., 0] += width / 2

    positions = zeros((0, 3))
    uvs = zeros((0, 2))
    normals = zeros((0, 3))

    faces = zeros((0, 3), dtype=dtype_indexes)
    outline = zeros((0, 2), dtype=dtype_indexes)

    offset = 0
    for vertices_p, faces_p, outline_p in planes_p:
        positions = np.vstack([positions, vertices_p["position"]])
        uvs = np.vstack([uvs, vertices_p["uv"]])
        normals = np.vstack([normals, vertices_p["normal"]])

        faces = np.vstack([faces, faces_p + offset])
        outline = np.vstack([outline, outline_p + offset])
        offset += vertices_p["position"].shape[0]

    vertices = zeros(
        positions.shape[0],
        [
            ("position", dtype_vertices, 3),
            ("uv", dtype_vertices, 2),
            ("normal", dtype_vertices, 3),
            ("colour", dtype_vertices, 4),
        ],  # type: ignore[arg-type]
    )

    vertex_colours = np.ravel(positions)
    vertex_colours = np.hstack(
        [
            np.reshape(
                np.interp(
                    vertex_colours,
                    (np.min(vertex_colours), np.max(vertex_colours)),
                    (0, 1),
                ),
                positions.shape,
            ),
            ones((positions.shape[0], 1)),
        ]
    )

    vertices["position"] = positions
    vertices["uv"] = uvs
    vertices["normal"] = normals
    vertices["colour"] = vertex_colours

    return vertices, faces, outline


PRIMITIVE_METHODS: CaseInsensitiveMapping = CaseInsensitiveMapping(
    {
        "Grid": primitive_grid,
        "Cube": primitive_cube,
    }
)
PRIMITIVE_METHODS.__doc__ = """
Supported geometry primitive generation methods.
"""


def primitive(
    method: Union[Literal["Cube", "Grid"], str] = "Cube", **kwargs: Any
) -> Tuple[NDArray, NDArray, NDArray]:
    """
    Return a geometry primitive using given method.

    Parameters
    ----------
    method
        Generation method.

    Other Parameters
    ----------------
    axis
        {:func:`colour.geometry.primitive_grid`},
        Axis the primitive will be normal to, or plane the primitive will be
        co-planar with.
    depth
        {:func:`colour.geometry.primitive_grid`,
        :func:`colour.geometry.primitive_cube`},
        Primitive depth.
    depth_segments
        {:func:`colour.geometry.primitive_grid`,
        :func:`colour.geometry.primitive_cube`},
        Primitive segments count along the depth.
    dtype_indexes
        {:func:`colour.geometry.primitive_grid`,
        :func:`colour.geometry.primitive_cube`},
        :class:`numpy.dtype` to use for the grid indexes, default to
        the :class:`numpy.dtype` defined by the
        :attr:`colour.constant.DEFAULT_INT_DTYPE` attribute.
    dtype_vertices
        {:func:`colour.geometry.primitive_grid`,
        :func:`colour.geometry.primitive_cube`},
        :class:`numpy.dtype` to use for the grid vertices, default to
        the :class:`numpy.dtype` defined by the
        :attr:`colour.constant.DEFAULT_FLOAT_DTYPE` attribute.
    height
        {:func:`colour.geometry.primitive_grid`,
        :func:`colour.geometry.primitive_cube`},
        Primitive height.
    planes
        {:func:`colour.geometry.primitive_cube`},
        Included grid primitives in the cube construction.
    width
        {:func:`colour.geometry.primitive_grid`,
        :func:`colour.geometry.primitive_cube`},
        Primitive width.
    width_segments
        {:func:`colour.geometry.primitive_grid`,
        :func:`colour.geometry.primitive_cube`},
        Primitive segments count along the width.
    height_segments
        {:func:`colour.geometry.primitive_grid`,
        :func:`colour.geometry.primitive_cube`},
        Primitive segments count along the height.

    Returns
    -------
    :class:`tuple`
        Tuple of primitive vertices, face indexes to produce a filled primitive
        and outline indexes to produce an outline of the faces of the
        primitive.

    References
    ----------
    :cite:`Cabello2015`

    Examples
    --------
    >>> vertices, faces, outline = primitive()
    >>> print(vertices)
    [([-0.5,  0.5, -0.5], [ 0.,  1.], [-0., -0., -1.], [ 0.,  1.,  0.,  1.])
     ([ 0.5,  0.5, -0.5], [ 1.,  1.], [-0., -0., -1.], [ 1.,  1.,  0.,  1.])
     ([-0.5, -0.5, -0.5], [ 0.,  0.], [-0., -0., -1.], [ 0.,  0.,  0.,  1.])
     ([ 0.5, -0.5, -0.5], [ 1.,  0.], [-0., -0., -1.], [ 1.,  0.,  0.,  1.])
     ([-0.5,  0.5,  0.5], [ 0.,  1.], [ 0.,  0.,  1.], [ 0.,  1.,  1.,  1.])
     ([ 0.5,  0.5,  0.5], [ 1.,  1.], [ 0.,  0.,  1.], [ 1.,  1.,  1.,  1.])
     ([-0.5, -0.5,  0.5], [ 0.,  0.], [ 0.,  0.,  1.], [ 0.,  0.,  1.,  1.])
     ([ 0.5, -0.5,  0.5], [ 1.,  0.], [ 0.,  0.,  1.], [ 1.,  0.,  1.,  1.])
     ([ 0.5, -0.5, -0.5], [ 0.,  1.], [-0., -1., -0.], [ 1.,  0.,  0.,  1.])
     ([ 0.5, -0.5,  0.5], [ 1.,  1.], [-0., -1., -0.], [ 1.,  0.,  1.,  1.])
     ([-0.5, -0.5, -0.5], [ 0.,  0.], [-0., -1., -0.], [ 0.,  0.,  0.,  1.])
     ([-0.5, -0.5,  0.5], [ 1.,  0.], [-0., -1., -0.], [ 0.,  0.,  1.,  1.])
     ([ 0.5,  0.5, -0.5], [ 0.,  1.], [ 0.,  1.,  0.], [ 1.,  1.,  0.,  1.])
     ([ 0.5,  0.5,  0.5], [ 1.,  1.], [ 0.,  1.,  0.], [ 1.,  1.,  1.,  1.])
     ([-0.5,  0.5, -0.5], [ 0.,  0.], [ 0.,  1.,  0.], [ 0.,  1.,  0.,  1.])
     ([-0.5,  0.5,  0.5], [ 1.,  0.], [ 0.,  1.,  0.], [ 0.,  1.,  1.,  1.])
     ([-0.5, -0.5,  0.5], [ 0.,  1.], [-1., -0., -0.], [ 0.,  0.,  1.,  1.])
     ([-0.5,  0.5,  0.5], [ 1.,  1.], [-1., -0., -0.], [ 0.,  1.,  1.,  1.])
     ([-0.5, -0.5, -0.5], [ 0.,  0.], [-1., -0., -0.], [ 0.,  0.,  0.,  1.])
     ([-0.5,  0.5, -0.5], [ 1.,  0.], [-1., -0., -0.], [ 0.,  1.,  0.,  1.])
     ([ 0.5, -0.5,  0.5], [ 0.,  1.], [ 1.,  0.,  0.], [ 1.,  0.,  1.,  1.])
     ([ 0.5,  0.5,  0.5], [ 1.,  1.], [ 1.,  0.,  0.], [ 1.,  1.,  1.,  1.])
     ([ 0.5, -0.5, -0.5], [ 0.,  0.], [ 1.,  0.,  0.], [ 1.,  0.,  0.,  1.])
     ([ 0.5,  0.5, -0.5], [ 1.,  0.], [ 1.,  0.,  0.], [ 1.,  1.,  0.,  1.])]
    >>> print(faces)
    [[ 1  2  0]
     [ 1  3  2]
     [ 4  6  5]
     [ 6  7  5]
     [ 9 10  8]
     [ 9 11 10]
     [12 14 13]
     [14 15 13]
     [17 18 16]
     [17 19 18]
     [20 22 21]
     [22 23 21]]
    >>> print(outline)
    [[ 0  2]
     [ 2  3]
     [ 3  1]
     [ 1  0]
     [ 4  6]
     [ 6  7]
     [ 7  5]
     [ 5  4]
     [ 8 10]
     [10 11]
     [11  9]
     [ 9  8]
     [12 14]
     [14 15]
     [15 13]
     [13 12]
     [16 18]
     [18 19]
     [19 17]
     [17 16]
     [20 22]
     [22 23]
     [23 21]
     [21 20]]
    >>> vertices, faces, outline = primitive('Grid')
    >>> print(vertices)
    [([-0.5,  0.5,  0. ], [ 0.,  1.], [ 0.,  0.,  1.], [ 0.,  1.,  0.,  1.])
     ([ 0.5,  0.5,  0. ], [ 1.,  1.], [ 0.,  0.,  1.], [ 1.,  1.,  0.,  1.])
     ([-0.5, -0.5,  0. ], [ 0.,  0.], [ 0.,  0.,  1.], [ 0.,  0.,  0.,  1.])
     ([ 0.5, -0.5,  0. ], [ 1.,  0.], [ 0.,  0.,  1.], [ 1.,  0.,  0.,  1.])]
    >>> print(faces)
    [[0 2 1]
     [2 3 1]]
    >>> print(outline)
    [[0 2]
     [2 3]
     [3 1]
     [1 0]]
    """

    method = validate_method(method, PRIMITIVE_METHODS)

    function = PRIMITIVE_METHODS[method]

    return function(**filter_kwargs(function, **kwargs))



from __future__ import annotations

from colour.colorimetry import SpectralDistribution
from colour.hints import Floating, Literal, Union

from .datasets import *  # noqa
from . import datasets
from .cfi2017 import (
    ColourRendering_Specification_CIE2017,
    colour_fidelity_index_CIE2017,
)
from .cri import ColourRendering_Specification_CRI, colour_rendering_index
from .cqs import (
    COLOUR_QUALITY_SCALE_METHODS,
    ColourRendering_Specification_CQS,
    colour_quality_scale,
)
from .ssi import spectral_similarity_index
from .tm3018 import (
    ColourQuality_Specification_ANSIIESTM3018,
    colour_fidelity_index_ANSIIESTM3018,
)
from colour.utilities import CaseInsensitiveMapping, validate_method

__all__ = []
__all__ += datasets.__all__
__all__ += [
    "ColourRendering_Specification_CIE2017",
    "colour_fidelity_index_CIE2017",
]
__all__ += [
    "ColourQuality_Specification_ANSIIESTM3018",
    "colour_fidelity_index_ANSIIESTM3018",
]
__all__ += [
    "ColourRendering_Specification_CRI",
    "colour_rendering_index",
]
__all__ += [
    "ColourRendering_Specification_CQS",
    "COLOUR_QUALITY_SCALE_METHODS",
    "colour_quality_scale",
]
__all__ += [
    "spectral_similarity_index",
]

COLOUR_FIDELITY_INDEX_METHODS = CaseInsensitiveMapping(
    {
        "CIE 2017": colour_fidelity_index_CIE2017,
        "ANSI/IES TM-30-18": colour_fidelity_index_ANSIIESTM3018,
    }
)
COLOUR_FIDELITY_INDEX_METHODS.__doc__ = """
Supported *Colour Fidelity Index* (CFI) computation methods.

References
----------
:cite:`CIETC1-902017`, :cite:`ANSI2018`
"""


def colour_fidelity_index(
    sd_test: SpectralDistribution,
    additional_data=False,
    method: Union[Literal["CIE 2017", "ANSI/IES TM-30-18"], str] = "CIE 2017",
) -> Union[
    Floating,
    ColourRendering_Specification_CIE2017,
    ColourQuality_Specification_ANSIIESTM3018,
]:
    """
    Return the *Colour Fidelity Index* (CFI) :math:`R_f` of given spectral
    distribution using given method.

    Parameters
    ----------
    sd_test
        Test spectral distribution.
    additional_data
        Whether to output additional data.
    method
        Computation method.

    Returns
    -------
    :class:`numpy.floating` or \
:class:`colour.quality.ColourRendering_Specification_CIE2017` or \
:class:`colour.quality.ColourQuality_Specification_ANSIIESTM3018`
        *Colour Fidelity Index* (CFI) :math:`R_f`.

    References
    ----------
    :cite:`CIETC1-902017`, :cite:`ANSI2018`

    Examples
    --------
    >>> from colour.colorimetry import SDS_ILLUMINANTS
    >>> sd = SDS_ILLUMINANTS['FL2']
    >>> colour_fidelity_index(sd)  # doctest: +ELLIPSIS
    70.1208254...
    """

    method = validate_method(method, COLOUR_FIDELITY_INDEX_METHODS)

    function = COLOUR_FIDELITY_INDEX_METHODS[method]

    return function(sd_test, additional_data)


__all__ += [
    "COLOUR_FIDELITY_INDEX_METHODS",
    "colour_fidelity_index",
]



"""
Otsu, Yamamoto and Hachisuka (2018) - Reflectance Recovery
==========================================================

Defines the objects for reflectance recovery, i.e. spectral upsampling, using
*Otsu et al. (2018)* method:

-   :class:`colour.recovery.Dataset_Otsu2018`
-   :func:`colour.recovery.XYZ_to_sd_Otsu2018`
-   :func:`colour.recovery.Tree_Otsu2018`

References
----------
-   :cite:`Otsu2018` : Otsu, H., Yamamoto, M., & Hachisuka, T. (2018).
    Reproducing Spectral Reflectances From Tristimulus Colours. Computer
    Graphics Forum, 37(6), 370-381. doi:10.1111/cgf.13332
"""

from __future__ import annotations

import numpy as np
from dataclasses import dataclass

from colour.colorimetry import (
    MultiSpectralDistributions,
    SpectralDistribution,
    SpectralShape,
    handle_spectral_arguments,
    msds_to_XYZ_integration,
    reshape_msds,
    sd_to_XYZ,
)
from colour.hints import (
    ArrayLike,
    Boolean,
    Callable,
    Dict,
    Floating,
    Integer,
    List,
    NDArray,
    Optional,
    Tuple,
    cast,
)
from colour.models import XYZ_to_xy
from colour.recovery import (
    SPECTRAL_SHAPE_OTSU2018,
    BASIS_FUNCTIONS_OTSU2018,
    CLUSTER_MEANS_OTSU2018,
    SELECTOR_ARRAY_OTSU2018,
)
from colour.utilities import (
    Node,
    as_float_array,
    domain_range_scale,
    is_tqdm_installed,
    message_box,
    optional,
    to_domain_1,
    zeros,
)

if is_tqdm_installed():
    from tqdm import tqdm
else:  # pragma: no cover
    from unittest import mock

    tqdm = mock.MagicMock()

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "Dataset_Otsu2018",
    "DATASET_REFERENCE_OTSU2018",
    "XYZ_to_sd_Otsu2018",
    "PartitionAxis",
    "Data_Otsu2018",
    "Node_Otsu2018",
    "Tree_Otsu2018",
]


class Dataset_Otsu2018:
    """
    Store all the information needed for the *Otsu et al. (2018)* spectral
    upsampling method.

    Datasets can be either generated and converted as a
    :class:`colour.recovery.Dataset_Otsu2018` class instance using the
    :meth:`colour.recovery.Tree_Otsu2018.to_dataset` method or
    alternatively, loaded from disk with the
    :meth:`colour.recovery.Dataset_Otsu2018.read` method.

    Parameters
    ----------
    shape
        Shape of the spectral data.
    basis_functions
        Three basis functions for every cluster.
    means
        Mean for every cluster.
    selector_array
        Array describing how to select the appropriate cluster. See
        :meth:`colour.recovery.Dataset_Otsu2018.select` method for details.

    Attributes
    ----------
    -   :attr:`~colour.recovery.Dataset_Otsu2018.shape`
    -   :attr:`~colour.recovery.Dataset_Otsu2018.basis_functions`
    -   :attr:`~colour.recovery.Dataset_Otsu2018.means`
    -   :attr:`~colour.recovery.Dataset_Otsu2018.selector_array`

    Methods
    -------
    -   :meth:`~colour.recovery.Dataset_Otsu2018.__init__`
    -   :meth:`~colour.recovery.Dataset_Otsu2018.select`
    -   :meth:`~colour.recovery.Dataset_Otsu2018.cluster`
    -   :meth:`~colour.recovery.Dataset_Otsu2018.read`
    -   :meth:`~colour.recovery.Dataset_Otsu2018.write`

    References
    ----------
    :cite:`Otsu2018`

    Examples
    --------
    >>> import os
    >>> import colour
    >>> from colour.characterisation import SDS_COLOURCHECKERS
    >>> from colour.colorimetry import sds_and_msds_to_msds
    >>> reflectances = sds_and_msds_to_msds(
    ...     SDS_COLOURCHECKERS['ColorChecker N Ohta'].values()
    ... )
    >>> node_tree = Tree_Otsu2018(reflectances)
    >>> node_tree.optimise(iterations=2, print_callable=lambda x: x)
    >>> dataset = node_tree.to_dataset()
    >>> path = os.path.join(colour.__path__[0], 'recovery', 'tests',
    ...                     'resources', 'ColorChecker_Otsu2018.npz')
    >>> dataset.write(path) # doctest: +SKIP
    >>> dataset = Dataset_Otsu2018() # doctest: +SKIP
    >>> dataset.read(path) # doctest: +SKIP
    """

    def __init__(
        self,
        shape: Optional[SpectralShape] = None,
        basis_functions: Optional[NDArray] = None,
        means: Optional[NDArray] = None,
        selector_array: Optional[NDArray] = None,
    ):
        self._shape: Optional[SpectralShape] = shape
        self._basis_functions: Optional[NDArray] = (
            basis_functions
            if basis_functions is None
            else as_float_array(basis_functions)
        )
        self._means: Optional[NDArray] = (
            means if means is None else as_float_array(means)
        )
        self._selector_array: Optional[NDArray] = (
            selector_array
            if selector_array is None
            else as_float_array(selector_array)
        )

    @property
    def shape(self) -> Optional[SpectralShape]:
        """
        Getter property for the shape used by the *Otsu et al. (2018)* dataset.

        Returns
        -------
        :py:data:`None` or :class:`colour.SpectralShape`
            Shape used by the *Otsu et al. (2018)* dataset.
        """

        return self._shape

    @property
    def basis_functions(self) -> Optional[NDArray]:
        """
        Getter property for the basis functions of the *Otsu et al. (2018)*
        dataset.

        Returns
        -------
        :py:data:`None` or :class:`numpy.ndarray`
            Basis functions of the *Otsu et al. (2018)* dataset.
        """

        return self._basis_functions

    @property
    def means(self) -> Optional[NDArray]:
        """
        Getter property for means of the *Otsu et al. (2018)* dataset.

        Returns
        -------
        :py:data:`None` or :class:`numpy.ndarray`
            Means of the *Otsu et al. (2018)* dataset.
        """

        return self._means

    @property
    def selector_array(self) -> Optional[NDArray]:
        """
        Getter property for the selector array of the *Otsu et al. (2018)*
        dataset.

        Returns
        -------
        :py:data:`None` or :class:`numpy.ndarray`
            Selector array of the *Otsu et al. (2018)* dataset.
        """

        return self._selector_array

    def __str__(self) -> str:
        """
        Return a formatted string representation of the dataset.

        Returns
        -------
        :class:`str`
            Formatted string representation.
        """

        if self._basis_functions is not None:
            return (
                f"{self.__class__.__name__}"
                f"({self._basis_functions.shape[0]} basis functions)"
            )
        else:
            return f"{self.__class__.__name__}()"

    def select(self, xy: ArrayLike) -> Integer:
        """
        Return the cluster index appropriate for the given *CIE xy*
        coordinates.

        Parameters
        ----------
        xy
            *CIE xy* chromaticity coordinates.

        Returns
        -------
        :class:`numpy.integer`
            Cluster index.

        Raises
        ------
        ValueError
            If the selector array is undefined.
        """

        xy = as_float_array(xy)

        if self._selector_array is not None:
            i = 0
            while True:
                row = self._selector_array[i, :]
                origin, direction, lesser_index, greater_index = row

                if xy[int(direction)] <= origin:
                    index = int(lesser_index)
                else:
                    index = int(greater_index)

                if index < 0:
                    i = -index
                else:
                    return index
        else:
            raise ValueError('The "selector array" is undefined!')

    def cluster(self, xy: ArrayLike) -> Tuple[NDArray, NDArray]:
        """
        Return the basis functions and dataset mean for the given *CIE xy*
        coordinates.

        Parameters
        ----------
        xy
            *CIE xy* chromaticity coordinates.

        Returns
        -------
        :class:`tuple`
            Tuple of three basis functions and dataset mean.

        Raises
        ------
        ValueError
            If the basis functions or means are undefined.
        """

        if self._basis_functions is not None and self._means is not None:
            index = self.select(xy)

            return self._basis_functions[index, :, :], self._means[index, :]
        else:
            raise ValueError('The "basis functions" or "means" are undefined!')

    def read(self, path: str):
        """
        Read and loads a dataset from an *.npz* file.

        Parameters
        ----------
        path
            Path to the file.

        Examples
        --------
        >>> import os
        >>> import colour
        >>> from colour.characterisation import SDS_COLOURCHECKERS
        >>> from colour.colorimetry import sds_and_msds_to_msds
        >>> reflectances = sds_and_msds_to_msds(
        ...     SDS_COLOURCHECKERS['ColorChecker N Ohta'].values()
        ... )
        >>> node_tree = Tree_Otsu2018(reflectances)
        >>> node_tree.optimise(iterations=2, print_callable=lambda x: x)
        >>> dataset = node_tree.to_dataset()
        >>> path = os.path.join(colour.__path__[0], 'recovery', 'tests',
        ...                     'resources', 'ColorChecker_Otsu2018.npz')
        >>> dataset.write(path) # doctest: +SKIP
        >>> dataset = Dataset_Otsu2018() # doctest: +SKIP
        >>> dataset.read(path) # doctest: +SKIP
        """

        data = np.load(path)

        start, end, interval = data["shape"]
        self._shape = SpectralShape(start, end, interval)
        self._basis_functions = data["basis_functions"]
        self._means = data["means"]
        self._selector_array = data["selector_array"]

    def write(self, path: str):
        """
        Write the dataset to an *.npz* file at given path.

        Parameters
        ----------
        path
            Path to the file.

        Raises
        ------
        ValueError
            If the shape is undefined.

        Examples
        --------
        >>> import os
        >>> import colour
        >>> from colour.characterisation import SDS_COLOURCHECKERS
        >>> from colour.colorimetry import sds_and_msds_to_msds
        >>> reflectances = sds_and_msds_to_msds(
        ...     SDS_COLOURCHECKERS['ColorChecker N Ohta'].values()
        ... )
        >>> node_tree = Tree_Otsu2018(reflectances)
        >>> node_tree.optimise(iterations=2, print_callable=lambda x: x)
        >>> dataset = node_tree.to_dataset()
        >>> path = os.path.join(colour.__path__[0], 'recovery', 'tests',
        ...                     'resources', 'ColorChecker_Otsu2018.npz')
        >>> dataset.write(path) # doctest: +SKIP
        """

        if self._shape is not None:
            np.savez(
                path,
                shape=as_float_array(
                    [
                        self._shape.start,
                        self._shape.end,
                        self._shape.interval,
                    ]
                ),
                basis_functions=cast(NDArray, self._basis_functions),
                means=cast(NDArray, self._means),
                selector_array=cast(NDArray, self._selector_array),
            )
        else:
            raise ValueError('The "shape" is undefined!')


DATASET_REFERENCE_OTSU2018: Dataset_Otsu2018 = Dataset_Otsu2018(
    SPECTRAL_SHAPE_OTSU2018,
    BASIS_FUNCTIONS_OTSU2018,
    CLUSTER_MEANS_OTSU2018,
    SELECTOR_ARRAY_OTSU2018,
)
"""
Builtin *Otsu et al. (2018)* dataset as a
:class:`colour.recovery.Dataset_Otsu2018` class instance, usable by
:func:`colour.recovery.XYZ_to_sd_Otsu2018` definition among others.
"""


def XYZ_to_sd_Otsu2018(
    XYZ: ArrayLike,
    cmfs: Optional[MultiSpectralDistributions] = None,
    illuminant: Optional[SpectralDistribution] = None,
    dataset: Dataset_Otsu2018 = DATASET_REFERENCE_OTSU2018,
    clip: Boolean = True,
) -> SpectralDistribution:
    """
    Recover the spectral distribution of given *CIE XYZ* tristimulus values
    using *Otsu et al. (2018)* method.

    Parameters
    ----------
    XYZ
        *CIE XYZ* tristimulus values to recover the spectral distribution from.
    cmfs
        Standard observer colour matching functions, default to the
        *CIE 1931 2 Degree Standard Observer*.
    illuminant
        Illuminant spectral distribution, default to
        *CIE Standard Illuminant D65*.
    dataset
        Dataset to use for reconstruction. The default is to use the published
        data.
    clip
        If *True*, the default, values below zero and above unity in the
        recovered spectral distributions will be clipped. This ensures that the
        returned reflectance is physical and conserves energy, but will cause
        noticeable colour differences in case of very saturated colours.

    Returns
    -------
    :class:`colour.SpectralDistribution`
        Recovered spectral distribution. Its shape is always that of the
        :class:`colour.recovery.SPECTRAL_SHAPE_OTSU2018` class instance.

    Raises
    ------
    ValueError
        If the dataset shape is undefined.

    References
    ----------
    :cite:`Otsu2018`

    Examples
    --------
    >>> from colour import (
    ...     CCS_ILLUMINANTS, SDS_ILLUMINANTS, MSDS_CMFS, XYZ_to_sRGB)
    >>> from colour.colorimetry import sd_to_XYZ_integration
    >>> from colour.utilities import numpy_print_options
    >>> XYZ = np.array([0.20654008, 0.12197225, 0.05136952])
    >>> cmfs = (
    ...     MSDS_CMFS['CIE 1931 2 Degree Standard Observer'].
    ...     copy().align(SPECTRAL_SHAPE_OTSU2018)
    ... )
    >>> illuminant = SDS_ILLUMINANTS['D65'].copy().align(cmfs.shape)
    >>> sd = XYZ_to_sd_Otsu2018(XYZ, cmfs, illuminant)
    >>> with numpy_print_options(suppress=True):
    ...     sd  # doctest: +ELLIPSIS
    SpectralDistribution([[ 380.        ,    0.0601939...],
                          [ 390.        ,    0.0568063...],
                          [ 400.        ,    0.0517429...],
                          [ 410.        ,    0.0495841...],
                          [ 420.        ,    0.0502007...],
                          [ 430.        ,    0.0506489...],
                          [ 440.        ,    0.0510020...],
                          [ 450.        ,    0.0493782...],
                          [ 460.        ,    0.0468046...],
                          [ 470.        ,    0.0437132...],
                          [ 480.        ,    0.0416957...],
                          [ 490.        ,    0.0403783...],
                          [ 500.        ,    0.0405197...],
                          [ 510.        ,    0.0406031...],
                          [ 520.        ,    0.0416912...],
                          [ 530.        ,    0.0430956...],
                          [ 540.        ,    0.0444474...],
                          [ 550.        ,    0.0459336...],
                          [ 560.        ,    0.0507631...],
                          [ 570.        ,    0.0628967...],
                          [ 580.        ,    0.0844661...],
                          [ 590.        ,    0.1334277...],
                          [ 600.        ,    0.2262428...],
                          [ 610.        ,    0.3599330...],
                          [ 620.        ,    0.4885571...],
                          [ 630.        ,    0.5752546...],
                          [ 640.        ,    0.6193023...],
                          [ 650.        ,    0.6450744...],
                          [ 660.        ,    0.6610548...],
                          [ 670.        ,    0.6688673...],
                          [ 680.        ,    0.6795426...],
                          [ 690.        ,    0.6887933...],
                          [ 700.        ,    0.7003469...],
                          [ 710.        ,    0.7084128...],
                          [ 720.        ,    0.7154674...],
                          [ 730.        ,    0.7234334...]],
                         interpolator=SpragueInterpolator,
                         interpolator_kwargs={},
                         extrapolator=Extrapolator,
                         extrapolator_kwargs={...})
    >>> sd_to_XYZ_integration(sd, cmfs, illuminant) / 100  # doctest: +ELLIPSIS
    array([ 0.2065494...,  0.1219712...,  0.0514002...])
    """

    shape = dataset.shape
    if shape is not None:
        XYZ = to_domain_1(XYZ)

        cmfs, illuminant = handle_spectral_arguments(
            cmfs, illuminant, shape_default=SPECTRAL_SHAPE_OTSU2018
        )

        xy = XYZ_to_xy(XYZ)

        basis_functions, mean = dataset.cluster(xy)

        M = np.empty((3, 3))
        for i in range(3):
            sd = SpectralDistribution(basis_functions[i, :], shape.range())

            with domain_range_scale("ignore"):
                M[:, i] = sd_to_XYZ(sd, cmfs, illuminant) / 100

        M_inverse = np.linalg.inv(M)

        sd = SpectralDistribution(mean, shape.range())

        with domain_range_scale("ignore"):
            XYZ_mu = sd_to_XYZ(sd, cmfs, illuminant) / 100

        weights = np.dot(M_inverse, XYZ - XYZ_mu)
        recovered_sd = np.dot(weights, basis_functions) + mean

        recovered_sd = np.clip(recovered_sd, 0, 1) if clip else recovered_sd

        return SpectralDistribution(recovered_sd, shape.range())
    else:
        raise ValueError('The dataset "shape" is undefined!')


@dataclass
class PartitionAxis:
    """
    Represent a horizontal or vertical line, partitioning the 2D space in
    two half-planes.

    Parameters
    ----------
    origin
        The x coordinate of a vertical line or the y coordinate of a horizontal
        line.
    direction
        *0* if vertical, *1* if horizontal.

    Methods
    -------
    -   :meth:`~colour.recovery.otsu2018.PartitionAxis.__str__`
    """

    origin: Floating
    direction: Integer

    def __str__(self):
        """
        Return a formatted string representation of the partition axis.

        Returns
        -------
        :class:`str`
            Formatted string representation.
        """

        return (
            f"{self.__class__.__name__}"
            f"({'horizontal' if self.direction else 'vertical'} partition "
            f"at {'y' if self.direction else 'x'} = {self.origin})"
        )


class Data_Otsu2018:
    """
    Store the reference reflectances and derived information along with the
    methods to process them for a leaf :class:`colour.recovery.otsu2018.Node`
    class instance.

    This class also supports partitioning: Creating two smaller instances of
    :class:`colour.recovery.otsu2018.Data` class by splitting along an
    horizontal or a vertical axis on the *CIE xy* plane.

    Parameters
    ----------
    reflectances
        Reference reflectances of the *n* colours to be stored.
        The shape must match ``tree.shape`` with *m* points for each colour.
    cmfs
        Standard observer colour matching functions.
    illuminant
        Illuminant spectral distribution.

    Attributes
    ----------
    -   :attr:`~colour.recovery.otsu2018.Data.reflectances`
    -   :attr:`~colour.recovery.otsu2018.Data.cmfs`
    -   :attr:`~colour.recovery.otsu2018.Data.illuminant`
    -   :attr:`~colour.recovery.otsu2018.Data.basis_functions`
    -   :attr:`~colour.recovery.otsu2018.Data.mean`

    Methods
    -------
    -   :meth:`~colour.recovery.otsu2018.Data.__str__`
    -   :meth:`~colour.recovery.otsu2018.Data.__len__`
    -   :meth:`~colour.recovery.otsu2018.Data.origin`
    -   :meth:`~colour.recovery.otsu2018.Data.partition`
    -   :meth:`~colour.recovery.otsu2018.Data.PCA`
    -   :meth:`~colour.recovery.otsu2018.Data.reconstruct`
    -   :meth:`~colour.recovery.otsu2018.Data.reconstruction_error`
    """

    def __init__(
        self,
        reflectances: Optional[ArrayLike],
        cmfs: MultiSpectralDistributions,
        illuminant: SpectralDistribution,
    ):
        self._cmfs: MultiSpectralDistributions = cmfs
        self._illuminant: SpectralDistribution = illuminant

        self._XYZ: Optional[NDArray] = None
        self._xy: Optional[NDArray] = None
        # TODO: Remove pragma when https://github.com/python/mypy/issues/3004
        # is resolved.
        self._reflectances: NDArray = np.array([])
        self.reflectances = reflectances  # type: ignore[assignment]

        self._basis_functions: Optional[NDArray] = None
        self._mean: Optional[Floating] = None
        self._M: Optional[NDArray] = None
        self._XYZ_mu: Optional[NDArray] = None

        self._reconstruction_error: Optional[Floating] = None

    @property
    def reflectances(self) -> NDArray:
        """
        Getter and setter property for the reference reflectances.

        Parameters
        ----------
        value
            Value to set the reference reflectances with.

        Returns
        -------
        :class:`numpy.ndarray`
            Reference reflectances.
        """

        return self._reflectances

    @reflectances.setter
    def reflectances(self, value: ArrayLike):
        """Setter for the **self.reflectances** property."""

        if value is not None:
            self._reflectances = as_float_array(value)
            self._XYZ = (
                msds_to_XYZ_integration(
                    self._reflectances,
                    self._cmfs,
                    self._illuminant,
                    shape=self._cmfs.shape,
                )
                / 100
            )
            self._xy = XYZ_to_xy(self._XYZ)
        else:
            self._reflectances, self._XYZ, self._xy = None, None, None

    @property
    def cmfs(self) -> MultiSpectralDistributions:
        """
        Getter property for the standard observer colour matching functions.

        Returns
        -------
        :class:`colour.MultiSpectralDistributions`
            Standard observer colour matching functions.
        """

        return self._cmfs

    @property
    def illuminant(self) -> SpectralDistribution:
        """
        Getter property for the illuminant.

        Returns
        -------
        :class:`colour.SpectralDistribution`
            Illuminant.
        """

        return self._illuminant

    @property
    def basis_functions(self) -> Optional[NDArray]:
        """
        Getter property for the basis functions.

        Returns
        -------
        :class:`numpy.ndarray`
            Basis functions.
        """

        return self._basis_functions

    @property
    def mean(self) -> Optional[Floating]:
        """
        Getter property for the mean distribution.

        Returns
        -------
        :class:`numpy.floating`
            Mean distribution.
        """

        return self._mean

    def __str__(self) -> str:
        """
        Return a formatted string representation of the data.

        Returns
        -------
        :class:`str`
            Formatted string representation.
        """

        return f"{self.__class__.__name__}({len(self)} Reflectances)"

    def __len__(self) -> Integer:
        """
        Return the number of colours in the data.

        Returns
        -------
        :class:`numpy.integer`
            Number of colours in the data.
        """

        return self._reflectances.shape[0]

    def origin(self, i: Integer, direction: Integer) -> Floating:
        """
        Return the origin *CIE x* or *CIE y* chromaticity coordinate for given
        index and direction.

        Parameters
        ----------
        i
            Origin index.
        direction
            Origin direction.

        Returns
        -------
        :class:`numpy.floating`
            Origin *CIE x* or *CIE y* chromaticity coordinate.

        Raises
        ------
        ValueError
            If the chromaticity coordinates are undefined.
        """

        if self._xy is not None:
            return self._xy[i, direction]
        else:
            raise ValueError('The "chromaticity coordinates" are undefined!')

    def partition(
        self, axis: PartitionAxis
    ) -> Tuple[Data_Otsu2018, Data_Otsu2018]:
        """
        Partition the data using given partition axis.

        Parameters
        ----------
        axis
            Partition axis used to partition the data.

        Returns
        -------
        :class:`tuple`
            Tuple of left or lower part and right or upper part.

        Raises
        ------
        ValueError
            If the tristimulus values or chromaticity coordinates are
            undefined.
        """

        lesser = Data_Otsu2018(None, self._cmfs, self._illuminant)
        greater = Data_Otsu2018(None, self._cmfs, self._illuminant)

        if self._XYZ is not None and self._xy is not None:
            mask = self._xy[:, axis.direction] <= axis.origin

            lesser._reflectances = self._reflectances[mask, :]
            greater._reflectances = self._reflectances[~mask, :]

            lesser._XYZ = self._XYZ[mask, :]
            greater._XYZ = self._XYZ[~mask, :]

            lesser._xy = self._xy[mask, :]
            greater._xy = self._xy[~mask, :]

            return lesser, greater
        else:
            raise ValueError(
                'The "tristimulus values" or "chromaticity coordinates" are '
                "undefined!"
            )

    def PCA(self):
        """
        Perform the *Principal Component Analysis* (PCA) on the data and sets
        the relevant attributes accordingly.
        """

        if self._M is None:
            settings = {
                "cmfs": self._cmfs,
                "illuminant": self._illuminant,
                "shape": self._cmfs.shape,
            }

            self._mean = np.mean(self.reflectances, axis=0)
            self._XYZ_mu = (
                msds_to_XYZ_integration(self._mean, **settings) / 100
            )

            matrix_data = self.reflectances - self._mean
            matrix_covariance = np.dot(np.transpose(matrix_data), matrix_data)
            _eigenvalues, eigenvectors = np.linalg.eigh(matrix_covariance)
            self._basis_functions = np.transpose(eigenvectors[:, -3:])

            self._M = np.transpose(
                msds_to_XYZ_integration(self._basis_functions, **settings)
                / 100
            )

    def reconstruct(self, XYZ: ArrayLike) -> SpectralDistribution:
        """
        Reconstruct the reflectance for the given *CIE XYZ* tristimulus
        values.

        Parameters
        ----------
        XYZ
            *CIE XYZ* tristimulus values to recover the spectral distribution
            from.

        Returns
        -------
        :class:`colour.SpectralDistribution`
            Recovered spectral distribution.

        Raises
        ------
        ValueError
            If the matrix :math:`M`, the mean tristimulus values or the basis
            functions are undefined.
        """

        if (
            self._M is not None
            and self._XYZ_mu is not None
            and self._basis_functions is not None
        ):
            XYZ = as_float_array(XYZ)

            weights = np.dot(np.linalg.inv(self._M), XYZ - self._XYZ_mu)
            reflectance = np.dot(weights, self._basis_functions) + self._mean
            reflectance = np.clip(reflectance, 0, 1)

            return SpectralDistribution(reflectance, self._cmfs.wavelengths)
        else:
            raise ValueError(
                'The matrix "M", the "mean tristimulus values" or the '
                '"basis functions" are undefined!'
            )

    def reconstruction_error(self) -> Floating:
        """
        Return the reconstruction error of the data. The error is computed by
        reconstructing the reflectances for the reference *CIE XYZ* tristimulus
        values using PCA and, comparing the reconstructed reflectances against
        the reference reflectances.

        Returns
        -------
        :class:`numpy.floating`
            The reconstruction error for the data.

        Raises
        ------
        ValueError
            If the tristimulus values are undefined.

        Notes
        -----
        -   The reconstruction error is cached upon being computed and thus is
            only computed once per node.
        """

        if self._reconstruction_error is not None:
            return self._reconstruction_error

        if self._XYZ is not None:
            self.PCA()

            error = 0
            for i in range(len(self)):
                sd = self._reflectances[i, :]
                XYZ = self._XYZ[i, :]
                recovered_sd = self.reconstruct(XYZ)
                error += np.sum((sd - recovered_sd.values) ** 2)

            self._reconstruction_error = error

            return error
        else:
            raise ValueError('The "tristimulus values" are undefined!')


class Node_Otsu2018(Node):
    """
    Represent a node in a :meth:`colour.recovery.Tree_Otsu2018` class instance
    node tree.

    Parameters
    ----------
    parent
        Parent of the node.
    children
        Children of the node.
    data
        The colour data belonging to this node.

    Attributes
    ----------
    -   :attr:`~colour.recovery.otsu2018.Node.partition_axis`
    -   :attr:`~colour.recovery.otsu2018.Node.row`

    Methods
    -------
    -   :meth:`~colour.recovery.otsu2018.Node.__init__`
    -   :meth:`~colour.recovery.otsu2018.Node.split`
    -   :meth:`~colour.recovery.otsu2018.Node.minimise`
    -   :meth:`~colour.recovery.otsu2018.Node.leaf_reconstruction_error`
    -   :meth:`~colour.recovery.otsu2018.Node.branch_reconstruction_error`
    """

    def __init__(
        self,
        parent: Optional[Node_Otsu2018] = None,
        children: Optional[List] = None,
        data: Optional[Data_Otsu2018] = None,
    ):
        super().__init__(parent=parent, children=children, data=data)

        self._partition_axis: Optional[PartitionAxis] = None
        self._best_partition: Optional[
            Tuple[List[Node_Otsu2018], PartitionAxis, Floating]
        ] = None

    @property
    def partition_axis(self) -> Optional[PartitionAxis]:
        """
        Getter property for the node partition axis.

        Returns
        -------
        :class:`colour.recovery.otsu2018.PartitionAxis`
            Node partition axis.
        """

        return self._partition_axis

    @property
    def row(self) -> Tuple[Floating, Integer, Node_Otsu2018, Node_Otsu2018]:
        """
        Getter property for the node row for the selector array.

        Returns
        -------
        :class:`tuple`
            Node row for the selector array.

        Raises
        ------
        ValueError
            If the partition axis is undefined.
        """

        if self._partition_axis is not None:
            return (
                self._partition_axis.origin,
                self._partition_axis.direction,
                cast(Node_Otsu2018, self.children[0]),
                cast(Node_Otsu2018, self.children[1]),
            )
        else:
            raise ValueError('The "partition axis" is undefined!')

    def split(self, children: List[Node_Otsu2018], axis: PartitionAxis):
        """
        Convert the leaf node into an inner node using given children and
        partition axis.

        Parameters
        ----------
        children
            Tuple of two :class:`colour.recovery.otsu2018.Node` class
            instances.
        axis
            Partition axis.
        """

        self.data = None
        self.children = list(children)

        self._best_partition = None
        self._partition_axis = axis

    def minimise(
        self, minimum_cluster_size: Integer
    ) -> Tuple[List[Node_Otsu2018], PartitionAxis, Floating]:
        """
        Find the best partition for the node that minimises the leaf
        reconstruction error.

        Parameters
        ----------
        minimum_cluster_size
            Smallest acceptable cluster size. It must be at least 3 or the
            *Principal Component Analysis* (PCA) is not be possible.

        Returns
        -------
        :class:`tuple`
            Tuple of tuple of nodes created by splitting a node with a given
            partition, partition axis, i.e. the horizontal or vertical line,
            partitioning the 2D space in two half-planes and partition error.
        """

        if self._best_partition is not None:
            return self._best_partition

        leaf_error = self.leaf_reconstruction_error()
        best_error = None

        with tqdm(total=2 * len(self.data)) as progress:
            for direction in [0, 1]:
                for i in range(len(self.data)):
                    progress.update()

                    axis = PartitionAxis(
                        self.data.origin(i, direction), direction
                    )
                    data_lesser, data_greater = self.data.partition(axis)

                    if np.any(
                        np.array(
                            [
                                len(data_lesser),
                                len(data_greater),
                            ]
                        )
                        < minimum_cluster_size
                    ):
                        continue

                    lesser = Node_Otsu2018(data=data_lesser)
                    lesser.data.PCA()

                    greater = Node_Otsu2018(data=data_greater)
                    greater.data.PCA()

                    partition_error = (
                        lesser.leaf_reconstruction_error()
                        + greater.leaf_reconstruction_error()
                    )

                    partition = [lesser, greater]

                    if partition_error >= leaf_error:
                        continue

                    if best_error is None or partition_error < best_error:
                        self._best_partition = (
                            partition,
                            axis,
                            partition_error,
                        )

        if self._best_partition is None:
            raise RuntimeError("Could not find the best partition!")

        return self._best_partition

    def leaf_reconstruction_error(self) -> Floating:
        """
        Return the reconstruction error of the node data. The error is
        computed by reconstructing the reflectances for the data reference
        *CIE XYZ* tristimulus values using PCA and, comparing the reconstructed
        reflectances against the data reference reflectances.

        Returns
        -------
        :class:`numpy.floating`
            The reconstruction errors summation for the node data.
        """

        return self.data.reconstruction_error()

    def branch_reconstruction_error(self) -> Floating:
        """
        Compute the reconstruction error for all the leaves data connected to
        the node or its children, i.e. the reconstruction errors summation for
        all the leaves in the branch.

        Returns
        -------
        :class:`numpy.floating`
            Reconstruction errors summation for all the leaves' data in the
            branch.
        """

        if self.is_leaf():
            return self.leaf_reconstruction_error()
        else:
            return np.sum(
                as_float_array(
                    [
                        cast(
                            Node_Otsu2018, child
                        ).branch_reconstruction_error()
                        for child in self.children
                    ]
                )
            )


class Tree_Otsu2018(Node_Otsu2018):
    """
    A sub-class of :class:`colour.recovery.otsu2018.Node` class representing
    the root node of a tree containing information shared with all the nodes,
    such as the standard observer colour matching functions and the illuminant,
    if any is used.

    Global operations involving the entire tree, such as optimisation and
    conversion to dataset, are implemented in this sub-class.

    Parameters
    ----------
    reflectances
        Reference reflectances of the *n* reference colours to use for
        optimisation.
    cmfs
        Standard observer colour matching functions, default to the
        *CIE 1931 2 Degree Standard Observer*.
    illuminant
        Illuminant spectral distribution, default to
        *CIE Standard Illuminant D65*.

    Attributes
    ----------
    -   :attr:`~colour.recovery.Tree_Otsu2018.reflectances`
    -   :attr:`~colour.recovery.Tree_Otsu2018.cmfs`
    -   :attr:`~colour.recovery.Tree_Otsu2018.illuminant`

    Methods
    -------
    -   :meth:`~colour.recovery.otsu2018.Tree_Otsu2018.__init__`
    -   :meth:`~colour.recovery.otsu2018.Tree_Otsu2018.__str__`
    -   :meth:`~colour.recovery.otsu2018.Tree_Otsu2018.optimise`
    -   :meth:`~colour.recovery.otsu2018.Tree_Otsu2018.to_dataset`

    References
    ----------
    :cite:`Otsu2018`

    Examples
    --------
    >>> import os
    >>> import colour
    >>> from colour import MSDS_CMFS, SDS_COLOURCHECKERS, SDS_ILLUMINANTS
    >>> from colour.colorimetry import sds_and_msds_to_msds
    >>> from colour.utilities import numpy_print_options
    >>> XYZ = np.array([0.20654008, 0.12197225, 0.05136952])
    >>> cmfs = (
    ...     MSDS_CMFS['CIE 1931 2 Degree Standard Observer'].
    ...     copy().align(SpectralShape(360, 780, 10))
    ... )
    >>> illuminant = SDS_ILLUMINANTS['D65'].copy().align(cmfs.shape)
    >>> reflectances = sds_and_msds_to_msds(
    ...     SDS_COLOURCHECKERS['ColorChecker N Ohta'].values()
    ... )
    >>> node_tree = Tree_Otsu2018(reflectances, cmfs, illuminant)
    >>> node_tree.optimise(iterations=2, print_callable=lambda x: x)
    >>> dataset = node_tree.to_dataset()
    >>> path = os.path.join(colour.__path__[0], 'recovery', 'tests',
    ...                     'resources', 'ColorChecker_Otsu2018.npz')
    >>> dataset.write(path) # doctest: +SKIP
    >>> dataset = Dataset_Otsu2018() # doctest: +SKIP
    >>> dataset.read(path) # doctest: +SKIP
    >>> sd = XYZ_to_sd_Otsu2018(XYZ, cmfs, illuminant, dataset)
    >>> with numpy_print_options(suppress=True):
    ...     sd  # doctest: +ELLIPSIS
    SpectralDistribution([[ 360.        ,    0.0651341...],
                          [ 370.        ,    0.0651341...],
                          [ 380.        ,    0.0651341...],
                          [ 390.        ,    0.0749684...],
                          [ 400.        ,    0.0815578...],
                          [ 410.        ,    0.0776439...],
                          [ 420.        ,    0.0721897...],
                          [ 430.        ,    0.0649064...],
                          [ 440.        ,    0.0567185...],
                          [ 450.        ,    0.0484685...],
                          [ 460.        ,    0.0409768...],
                          [ 470.        ,    0.0358964...],
                          [ 480.        ,    0.0307857...],
                          [ 490.        ,    0.0270148...],
                          [ 500.        ,    0.0273773...],
                          [ 510.        ,    0.0303157...],
                          [ 520.        ,    0.0331285...],
                          [ 530.        ,    0.0363027...],
                          [ 540.        ,    0.0425987...],
                          [ 550.        ,    0.0513442...],
                          [ 560.        ,    0.0579256...],
                          [ 570.        ,    0.0653850...],
                          [ 580.        ,    0.0929522...],
                          [ 590.        ,    0.1600326...],
                          [ 600.        ,    0.2586159...],
                          [ 610.        ,    0.3701242...],
                          [ 620.        ,    0.4702243...],
                          [ 630.        ,    0.5396261...],
                          [ 640.        ,    0.5737561...],
                          [ 650.        ,    0.590848 ...],
                          [ 660.        ,    0.5935371...],
                          [ 670.        ,    0.5923295...],
                          [ 680.        ,    0.5956326...],
                          [ 690.        ,    0.5982513...],
                          [ 700.        ,    0.6017904...],
                          [ 710.        ,    0.6016419...],
                          [ 720.        ,    0.5996892...],
                          [ 730.        ,    0.6000018...],
                          [ 740.        ,    0.5964443...],
                          [ 750.        ,    0.5868181...],
                          [ 760.        ,    0.5860973...],
                          [ 770.        ,    0.5614878...],
                          [ 780.        ,    0.5289331...]],
                         interpolator=SpragueInterpolator,
                         interpolator_kwargs={},
                         extrapolator=Extrapolator,
                         extrapolator_kwargs={...})
    """

    def __init__(
        self,
        reflectances: MultiSpectralDistributions,
        cmfs: Optional[MultiSpectralDistributions] = None,
        illuminant: Optional[SpectralDistribution] = None,
    ):
        super().__init__()

        cmfs, illuminant = handle_spectral_arguments(
            cmfs, illuminant, shape_default=SPECTRAL_SHAPE_OTSU2018
        )

        self._cmfs: MultiSpectralDistributions = cmfs
        self._illuminant: SpectralDistribution = illuminant

        self._reflectances: NDArray = np.transpose(
            reshape_msds(reflectances, self._cmfs.shape).values
        )

        self.data: Data_Otsu2018 = Data_Otsu2018(
            self._reflectances, self._cmfs, self._illuminant
        )

    @property
    def reflectances(self) -> NDArray:
        """
        Getter property for the reference reflectances.

        Returns
        -------
        :class:`numpy.ndarray`
            Reference reflectances.
        """

        return self._reflectances

    @property
    def cmfs(self) -> MultiSpectralDistributions:
        """
        Getter property for the standard observer colour matching functions.

        Returns
        -------
        :class:`colour.MultiSpectralDistributions`
            Standard observer colour matching functions.
        """

        return self._cmfs

    @property
    def illuminant(self) -> SpectralDistribution:
        """
        Getter property for the illuminant.

        Returns
        -------
        :class:`colour.SpectralDistribution`
            Illuminant.
        """

        return self._illuminant

    def optimise(
        self,
        iterations: Integer = 8,
        minimum_cluster_size: Optional[Integer] = None,
        print_callable: Callable = print,
    ):
        """
        Optimise the tree by repeatedly performing optimal partitioning of the
        nodes, creating a tree that minimises the total reconstruction error.

        Parameters
        ----------
        iterations
            Maximum number of splits. If the dataset is too small, this number
            might not be reached. The default is to create 8 clusters, like in
            :cite:`Otsu2018`.
        minimum_cluster_size
            Smallest acceptable cluster size. By default, it is chosen
            automatically, based on the size of the dataset and desired number
            of clusters. It must be at least 3 or the
            *Principal Component Analysis* (PCA) is not be possible.
        print_callable
            Callable used to print progress and diagnostic information.

        Examples
        --------
        >>> from colour.colorimetry import sds_and_msds_to_msds
        >>> from colour import MSDS_CMFS, SDS_COLOURCHECKERS, SDS_ILLUMINANTS
        >>> cmfs = (
        ...     MSDS_CMFS['CIE 1931 2 Degree Standard Observer']
        ...     .copy().align(SpectralShape(360, 780, 10))
        ... )
        >>> illuminant = SDS_ILLUMINANTS['D65'].copy().align(cmfs.shape)
        >>> reflectances = sds_and_msds_to_msds(
        ...     SDS_COLOURCHECKERS['ColorChecker N Ohta'].values()
        ... )
        >>> node_tree = Tree_Otsu2018(reflectances, cmfs, illuminant)
        >>> node_tree.optimise(iterations=2)  # doctest: +ELLIPSIS
        ======================================================================\
=========
        *                                                                     \
        *
        *   "Otsu et al. (2018)" Tree Optimisation                            \
        *
        *                                                                     \
        *
        ======================================================================\
=========
        Initial branch error is: 4.8705353...
        <BLANKLINE>
        Iteration 1 of 2:
        <BLANKLINE>
        Optimising "Tree_Otsu2018#...(Data_Otsu2018(24 Reflectances))"...
        <BLANKLINE>
        Splitting "Tree_Otsu2018#...(Data_Otsu2018(24 Reflectances))" into \
"Node_Otsu2018#...(Data_Otsu2018(10 Reflectances))" and \
"Node_Otsu2018#...(Data_Otsu2018(14 Reflectances))" along \
"PartitionAxis(horizontal partition at y = 0.3240945...)".
        Error is reduced by 0.0054840... and is now 4.8650513..., 99.9% of \
the initial error.
        <BLANKLINE>
        Iteration 2 of 2:
        <BLANKLINE>
        Optimising "Node_Otsu2018#...(Data_Otsu2018(10 Reflectances))"...
        Optimisation failed: Could not find the best partition!
        Optimising "Node_Otsu2018#...(Data_Otsu2018(14 Reflectances))"...
        <BLANKLINE>
        Splitting "Node_Otsu2018#...(Data_Otsu2018(14 Reflectances))" into \
"Node_Otsu2018#...(Data_Otsu2018(7 Reflectances))" and \
"Node_Otsu2018#...(Data_Otsu2018(7 Reflectances))" along \
"PartitionAxis(horizontal partition at y = 0.3600663...)".
        Error is reduced by 0.9681059... and is now 3.8969453..., 80.0% of \
the initial error.
        Tree optimisation is complete!
        >>> print(node_tree.render())  # doctest: +ELLIPSIS
        |----"Tree_Otsu2018#..."
            |----"Node_Otsu2018#..."
            |----"Node_Otsu2018#..."
                |----"Node_Otsu2018#..."
                |----"Node_Otsu2018#..."
        <BLANKLINE>
        >>> len(node_tree)
        4
        """

        default_cluster_size = len(self.data) / iterations // 2
        minimum_cluster_size = max(
            cast(int, optional(minimum_cluster_size, default_cluster_size)), 3
        )

        initial_branch_error = self.branch_reconstruction_error()

        message_box(
            '"Otsu et al. (2018)" Tree Optimisation',
            print_callable=print_callable,
        )

        print_callable(f"Initial branch error is: {initial_branch_error}")

        best_leaf, best_partition, best_axis, partition_error = [None] * 4

        for i in range(iterations):
            print_callable(f"\nIteration {i + 1} of {iterations}:\n")

            total_error = self.branch_reconstruction_error()
            optimised_total_error = None

            for leaf in self.leaves:
                print_callable(f'Optimising "{leaf}"...')

                try:
                    partition, axis, partition_error = leaf.minimise(
                        minimum_cluster_size
                    )
                except RuntimeError as error:
                    print_callable(f"Optimisation failed: {error}")
                    continue

                new_total_error = (
                    total_error
                    - leaf.leaf_reconstruction_error()
                    + partition_error
                )

                if (
                    optimised_total_error is None
                    or new_total_error < optimised_total_error
                ):
                    optimised_total_error = new_total_error
                    best_axis = axis
                    best_leaf = leaf
                    best_partition = partition

            if optimised_total_error is None:
                print_callable(
                    f"\nNo further improvement is possible!"
                    f"\nTerminating at iteration {i}.\n"
                )
                break

            if best_partition is not None:
                print_callable(
                    f'\nSplitting "{best_leaf}" into "{best_partition[0]}" '
                    f'and "{best_partition[1]}" along "{best_axis}".'
                )

            print_callable(
                f"Error is reduced by "
                f"{leaf.leaf_reconstruction_error() - partition_error} and "
                f"is now {optimised_total_error}, "
                f"{100 * optimised_total_error / initial_branch_error:.1f}% "
                f"of the initial error."
            )

            if best_leaf is not None:
                best_leaf.split(best_partition, best_axis)

        print_callable("Tree optimisation is complete!")

    def to_dataset(self) -> Dataset_Otsu2018:
        """
        Create a :class:`colour.recovery.Dataset_Otsu2018` class instance
        based on data stored in the tree.

        The dataset can then be saved to disk or used to recover reflectance
        with :func:`colour.recovery.XYZ_to_sd_Otsu2018` definition.

        Returns
        -------
        :class:`colour.recovery.Dataset_Otsu2018`
            The dataset object.

        Examples
        --------
        >>> from colour.colorimetry import sds_and_msds_to_msds
        >>> from colour.characterisation import SDS_COLOURCHECKERS
        >>> reflectances = sds_and_msds_to_msds(
        ...     SDS_COLOURCHECKERS['ColorChecker N Ohta'].values()
        ... )
        >>> node_tree = Tree_Otsu2018(reflectances)
        >>> node_tree.optimise(iterations=2, print_callable=lambda x: x)
        >>> node_tree.to_dataset()  # doctest: +ELLIPSIS
        <colour.recovery.otsu2018.Dataset_Otsu2018 object at 0x...>
        """

        basis_functions = as_float_array(
            [leaf.data.basis_functions for leaf in self.leaves]
        )

        means = as_float_array([leaf.data.mean for leaf in self.leaves])

        if len(self.children) == 0:
            selector_array = zeros(4)
        else:

            def add_rows(node, data=None):
                """Add rows for given node and its children."""

                data = cast(
                    Dict,
                    optional(
                        data, {"rows": [], "node_to_leaf_id": {}, "leaf_id": 0}
                    ),
                )

                if node.is_leaf():
                    data["node_to_leaf_id"][node] = data["leaf_id"]
                    data["leaf_id"] += 1
                    return

                data["node_to_leaf_id"][node] = -len(data["rows"])
                data["rows"].append(list(node.row))

                for child in node.children:
                    add_rows(child, data)

                return data

            data = add_rows(self)
            rows = data["rows"]

            for i, row in enumerate(rows):
                for j in (2, 3):
                    rows[i][j] = data["node_to_leaf_id"][row[j]]

            selector_array = as_float_array(rows)

        return Dataset_Otsu2018(
            self._cmfs.shape,
            basis_functions,
            means,
            selector_array,
        )
